# Copyright (c) 2014-2022 Matteo Degiacomi
#
# BiobOx is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# BiobOx is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with BiobOx ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
#
# Author : Matteo Degiacomi, matteo.degiacomi@gmail.com

import numpy as np
import pandas as pd

from biobox.classes.polyhedron import Polyhedron
from biobox.classes.molecule import Molecule


class Multimer(Polyhedron):
    '''
    Construct and manipulate a protein assembly composed of several :func:`Molecule <molecule.Molecule>` instances. Subclass of :func:`Polyhedron <polyhedron.Polyhedron>`.
    '''

    def query(self, query_text, get_index=False):
        '''
        ## select specific atoms in a multimer un the basis of a text query.

        :param query_text: string selecting atoms of interest. Uses the pandas query syntax, can access all columns in the dataframe self.data.
        :param get_index: if set to True, returns the indices of selected atoms in self.points array (and self.data)
        :returns: coordinates of the selected points (in a unique array) and, if get_index is set to true, a list of their indices in subunits' self.points array.
        '''

        idx = self.data.query(query_text).index.values

        res = self.data.iloc[idx] #this is a new sliced dataframe
        targets = np.array(res.loc[:, ["unit", "unit_index"]].values)

        # append the coordinates of every unit within the query
        pts = np.empty([0, 3])
        for u in np.unique(targets[:, 0]):
            pos = targets[targets[:, 0] == u, 1].astype(int)
            this_unit = self.unit_labels[u]
            pts = np.concatenate((pts, self.unit[this_unit].points[pos]))

        if get_index:
            return [pts, idx]
        else:
            return pts


    def atomselect(self, u, chain, resid, atom, get_index=False, use_resname=False):
        '''
        ## select specific atoms in a multimer providing unit, chain, residue ID and atom name.

        :param u: number of desired unit to select in the multimer
        :param chain: selection of a specific chain name (accepts '*' as wildcard). Can also be a list or numpy array of strings.
        :param resid: residue ID of desired atoms (accepts '*' as wildcard). Can also be a list or numpy array of of int.
        :param atom: name of desired atom (accepts '*' as wildcard). Can also be a list or numpy array of strings.
        :param get_index: if set to True, returns the indices of selected atoms in self.points array (and self.data)
        :param use_resname: if set to True, consider information in "res" variable as resnames, and not resids
        :returns: coordinates of the selected points (in a unique array) and, if get_index is set to true, a list of their indices in subunits' self.points array.
        '''

        # extract id of units of interest
        if u == '*':
            unit_id = list(self.unit_labels.values())
        else:
            if isinstance(u, str) or isinstance(u, int):
                try:
                    unit_id = [self.unit_labels[str(u)]]
                except Exception as ex:
                    raise Exception("ERROR: unit %s not found!" % u)

            elif isinstance(u, list) or type(u).__module__ == 'numpy':
                unit_id = []
                for c in range(0, len(u), 1):
                    try:
                        unit_id.append(self.unit_labels[str(u[c])])
                    except Exception as ex:
                        raise Exception("ERROR: unit %s not found!" % u[c])
            else:
                raise Exception("ERROR: wrong type for unit selection. Should be str, int, list, or numpy")

        # initialize storage for indices and coordinates
        indices = []
        pts = np.empty([0, 3])
        for i in range(0, len(self.unit), 1):
            if i in unit_id:
                [pts_tmp, index_tmp] = self.unit[i].atomselect(chain, resid, atom, True, use_resname=use_resname)
                pts = np.concatenate((pts, pts_tmp))
                indices.append(index_tmp)
            else:
                # indices of all units must be stored. If unit is not
                # requested, return an empty array for it
                indices.append([])
        if get_index:
            return [pts, indices]
        else:
            return pts

    def make_molecule(self, rename_chains=False):
        '''
        Return a :func:`Molecule <molecule.Molecule>` object containing all the points of the assembly. Chain will indicate different units, original chain value is pushed in segment entry.

        :param rename_chains: if True, chains of the newly produced molecule will be named from scratch.

        :returns: :func:`Molecule <molecule.Molecule>` object
        '''

        # create new data entry (renumber indices, reassign chain name)
        data = np.empty([0, 9])

        atom_ccs = {}
        r = []
        c = []
        skipcharge = False
        for i in range(0, len(self.unit), 1):
            data_tmp = self.unit[i].data[[
                "atom", "index", "name", "resname", "chain",
                "resid", "beta", "occupancy", "atomtype"]].values

            if rename_chains:
                data_tmp[:, 4] = self.chain_names[i] # avoid renaming the chains

            data = np.concatenate((data, data_tmp))

            # merge knowledge about CCS acquired by different molecules
            atom_ccs = {}
            for k in self.unit[i].knowledge['atom_ccs'].keys():
                atom_ccs[k] = self.unit[i].knowledge['atom_ccs'][k]

            if len(r) == 0:
                r = self.unit[i].data['radius']
            else:
                r = np.concatenate((r, self.unit[i].data['radius']))

            try:
                if len(c) == 0:
                    c = self.unit[i].data['charge']
                else:
                    c = np.concatenate((c, self.unit[i].data['charge']))
            except Exception as ex:
                skipcharge = True
                continue

        data[:, 1] = np.linspace(1, len(data), len(data)).astype(int)
        cols = ["atom", "index", "name", "resname", "chain", "resid", "beta", "occupancy", "atomtype"]
        idx = np.arange(len(data))

        # create molecule, and push created data information
        M = Molecule()
        M.add_xyz(self.get_all_xyz())
        M.data = pd.DataFrame(data, index=idx, columns=cols)
        M.properties['center'] = M.get_center()
        M.knowledge['atom_ccs'] = atom_ccs
        M.data['radius'] = r

        if not skipcharge:
            M.data['charge'] = c



        return M

    #def rmsd(self, ref_index, u="*", chain="*", resid="*", atom="*", align=False):
    #    '''
    #    Calculate the RMSD between atoms of interest in all structure with respect of a reference structure.

    #    supposes that all multimer subunits contain the same amount of alternative coordiantes.
    #    These are considered as representations of monomers conformations in a possible multimer.

    #    :param u: number of desired unit to select in the multimer
    #    :param chain: selection of a specific chain name (accepts '*' as wildcard). Can also be a list or numpy array of strings.
    #    :param resid: residue ID of desired atoms (accepts '*' as wildcard). Can also be a list or numpy array of of int.
    #    :param atom: name of desired atom (accepts '*' as wildcard). Can also be a list or numpy array of strings.
    #    :param ref_index: index of reference structure in conformations database
    #    :param align: if True, structures will all be aligned (cannot be undone)
    #    :returns: RMSD of all structures with respect of reference structure (in a numpy array)
    #    '''
    #
    #    if ref_index >= self.unit[0].coordinates.shape[0]:
    #        raise Exception("ERROR: requested frame %s as reference, but only %s frames are available!" %(ref_index, self.unit[0].coordinates.shape[0]))

    #    # select indices of atoms of interest and call overloaded method
    #    indices = self.atomselect(u, chain, resid, atom, get_index=True)[1]
    #    return super(Multimer, self).rmsd(ref_index, points_indices=indices, align=align)

    def get_data(self, indices, columns):
        '''
        Return information about atom of interest (i.e., slice the data DataFrame)

        :param indices: list of indices
        :param columns: list of columns (e.g. ["resname", "resid", "chain"])
        :returns: slice of molecule's data DataFrame
        '''

        return self.data.loc[indices, columns].values

    def write_pdb(self, outname, rename_chains=False):
        '''
        Write a pdb of the multimeric assembly.

        :param outname: name of PDB file to generate
        '''

        #M = self.make_molecule(rename_chains)
        #M.write_pdb(outname)

        f_out = open(outname, "w")

        for f in range(len(self.unit[0].coordinates)):

            cnt = 1
            # set current state to new frame
            for j in range(0, len(self.unit), 1):
                self.unit[j].set_current(f)

        for j in range(0, len(self.unit), 1):
            # get data about points and their properties from the desired
            # protein structure
            d = self.unit[j].get_pdb_data()

            for i in range(0, len(self.unit[j].points), 1):
                # create and write PDB lin
                if d[i][2][0].isdigit():
                    L = '%-6s%5i %-5s%-4s%1s%4i    %8.3f%8.3f%8.3f%6.2f%6.2f          %2s\n' % (d[i][0], cnt, d[i][2], d[i][3], self.chain_names[j], int(d[i][5]), float(d[i][6]), float(d[i][7]), float(d[i][8]), float(d[i][9]), float(d[i][10]), d[i][11])
                else:
                    L = '%-6s%5i  %-4s%-4s%1s%4i    %8.3f%8.3f%8.3f%6.2f%6.2f          %2s\n' % (d[i][0], cnt, d[i][2], d[i][3], self.chain_names[j], int(d[i][5]), float(d[i][6]), float(d[i][7]), float(d[i][8]), float(d[i][9]), float(d[i][10]), d[i][11])
                #L='%-6s%5i  %-4s%-4s%1s%4i    %8.3f%8.3f%8.3f%6.2f%6.2f          %2s\n'%(d[i][0], cnt, d[i][2], d[i][3], self.chain_names[j], d[i][5], d[i][6], d[i][7], d[i][8], d[i][9], d[i][10], d[i][11])
                f_out.write(L)
                cnt += 1

            f_out.write("TER\n")

        f_out.write("END\n")
        f_out.close()



# Read 'mrc' or 'ccp4' or 'imod' map file format electron microscope data.
# Byte swapping will be done if needed.
#
# NOTE: the code below is a Python 3 translation of CHIMERA files
# chimera/share/VolumeData/mrc, readarray.py and griddata.py (2014 version, in Python 2)

import numpy as np
import os.path
from functools import reduce, cmp_to_key
#from density import Structure

def cmp(a, b):
    return (a > b) - (a < b)

# -----------------------------------------------------------------------------
# Maintain a cache of data objects using a limited amount of memory.
# The least recently accessed data is released first.
# -----------------------------------------------------------------------------
#
class Data_Cache:

    def __init__(self, size):
        self.size = size
        self.used = 0
        self.time = 1
        self.data = {}
        self.groups = {}


    # ---------------------------------------------------------------------------
    #
    def cache_data(self, key, value, size, description, groups = []):

        self.remove_key(key)
        d = Cached_Data(key, value, size, description, self.time_stamp(), groups)
        self.data[key] = d

        for g in groups:
            gtable = self.groups
            if not g in gtable:#.has_key(g):
                gtable[g] = []
            gtable[g].append(d)

        self.used = self.used + size
        self.reduce_use()

    # ---------------------------------------------------------------------------
    #
    def lookup_data(self, key):

        data = self.data
        if key in data: #data.has_key(key):
            d = data[key]
            d.last_access = self.time_stamp()
            v = d.value
        else:
            v = None

        self.reduce_use()
        return v

    # ---------------------------------------------------------------------------
    #
    def remove_key(self, key):

        data = self.data
        if key in data: #.has_key(key):
            self.remove_data(data[key])
        self.reduce_use()

    # ---------------------------------------------------------------------------
    #
    def group_keys_and_data(self, group):

        groups = self.groups
        if not group in groups: #.has_key(group):
            return []

        kd = map(lambda d: (d.key, d.value), groups[group])
        return kd

    # ---------------------------------------------------------------------------
    #
    def resize(self, size):

        self.size = size
        self.reduce_use()

    # ---------------------------------------------------------------------------
    #
    def reduce_use(self):

        if self.used <= self.size:
            return

        data = self.data
        dlist = list(data.values())
        dlist.sort(key=cmp_to_key(lambda d1, d2: cmp(d1.last_access, d2.last_access)))
        import sys
        for d in dlist:
            if sys.getrefcount(d.value) == 2:
                self.remove_data(d)
                if self.used <= self.size:
                    break

    # ---------------------------------------------------------------------------
    #
    def remove_data(self, d):

        del self.data[d.key]
        self.used = self.used - d.size
        d.value = None

        for g in d.groups:
            dlist = self.groups[g]
            dlist.remove(d)
            if len(dlist) == 0:
                del self.groups[g]

    # ---------------------------------------------------------------------------
    #
    def time_stamp(self):

        t = self.time
        self.time = t + 1
        return t

###############################################################################

class Cached_Data:

    def __init__(self, key, value, size, description, time_stamp, groups):

        self.key = key
        self.value = value
        self.size = size
        self.description = description
        self.last_access = time_stamp
        self.groups = groups

###############################################################################

class MRC_Grid:

    def __init__(self, path, file_type = 'mrc'):

        d = MRC_Data(path, file_type)
        self.mrc_data = d

        # Path, file_type used for reloading data sets.
        self.path = path
        self.file_type = file_type    # 'mrc', 'ccp4', ....

        name = self.name_from_path(path)
        self.name = name

        self.size = tuple(d.data_size)

        if not isinstance(d.element_type, np.dtype):
            d.element_type = np.dtype(d.element_type)
        self.value_type = d.element_type                # numpy dtype.


        # Parameters defining how data matrix is positioned in space
        self.origin = tuple(d.data_origin)
        self.original_origin = self.origin
        self.step = tuple(d.data_step)
        self.original_step = self.step
        self.cell_angles = tuple(d.cell_angles)
        self.rotation = tuple(map(tuple, d.rotation))
        self.symmetries = ()
        self.ijk_to_xyz_transform = None
        self.xyz_to_ijk_transform = None

        #self.rgba = default_color                        # preferred color for displaying data

        global data_cache
        self.data_cache = data_cache

        self.writable = False
        self.change_callbacks = []

        self.update_transform()

    # ---------------------------------------------------------------------------
    #
    def name_from_path(self, path):

        if isinstance(path, (list,tuple)):
                p = path[0]
        else:
                p = path

        name = os.path.basename(p)
        return name

    # ---------------------------------------------------------------------------
    # Compute 3 by 4 matrices encoding rotation and translation.
    #
    def update_transform(self):

        #from Matrix import skew_axes
        #saxes = skew_axes(self.cell_angles)
        saxes=[[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]]
        rsaxes = [apply_rotation(self.rotation, a) for a in saxes]
        tf, tf_inv = transformation_and_inverse(self.origin, self.step, rsaxes)
        if tf != self.ijk_to_xyz_transform or tf_inv != self.xyz_to_ijk_transform:
            self.ijk_to_xyz_transform = tf
            self.xyz_to_ijk_transform = tf_inv
            self.coordinates_changed()

    # ---------------------------------------------------------------------------
    # A matrix ijk corresponds to a point in xyz space.
    # This function maps the xyz point to the matrix index.
    # The returned matrix index need not be integers.
    #
    def xyz_to_ijk(self, xyz):

        return map_point(xyz, self.xyz_to_ijk_transform)

    # ---------------------------------------------------------------------------
    # A matrix ijk corresponds to a point in xyz space.
    # This function maps the matrix index to the xyz point.
    #
    def ijk_to_xyz(self, ijk):

        return map_point(ijk, self.ijk_to_xyz_transform)

    # ---------------------------------------------------------------------------
    #
    def matrix(self, ijk_origin = (0,0,0), ijk_size = None,
                         ijk_step = (1,1,1), progress = None, from_cache_only = False):

        if ijk_size == None:
            ijk_size = self.size

        m = self.cached_data(ijk_origin, ijk_size, ijk_step)
        if m is None and not from_cache_only:
            m = self.read_matrix(ijk_origin, ijk_size, ijk_step, progress)
            self.cache_data(m, ijk_origin, ijk_size, ijk_step)

        return m

    # ---------------------------------------------------------------------------
    # NumPy matrix.    The returned matrix has size ijk_size and
    # element ijk is accessed as m[k,j,i].    It is an error if the requested
    # submatrix does not lie completely within the full data matrix.    It is
    # also an error for the size to be <= 0 in any dimension.    These invalid
    # inputs might throw an exception or might return garbage.    It is the
    # callers responsibility to make sure the arguments are valid.
    #
    def read_matrix(self, ijk_origin, ijk_size, ijk_step, progress):

        return self.mrc_data.read_matrix(ijk_origin, ijk_size, ijk_step, progress)


    def cached_data(self, origin, size, step):

        dcache = self.data_cache
        if dcache is None:
            return None

        key = (self, tuple(origin), tuple(size), tuple(step))
        m = dcache.lookup_data(key)
        if not m is None:
            return m

        # Look for a matrix containing the desired matrix
        group = self
        kd = dcache.group_keys_and_data(group)
        for k, matrix in kd:
            orig, sz, st = k[1:]
            if (step[0] < st[0] or step[1] < st[1] or step[2] < st[2] or
                    step[0] % st[0] or step[1] % st[1] or step[2] % st[2]):
                continue                # Step sizes not compatible
            if (origin[0] < orig[0] or origin[1] < orig[1] or origin[2] < orig[2] or
                    origin[0] + size[0] > orig[0] + sz[0] or
                    origin[1] + size[1] > orig[1] + sz[1] or
                    origin[2] + size[2] > orig[2] + sz[2]):
                continue                # Doesn't cover.
            dstep = map(lambda a,b: a/b, step, st)
            offset = map(lambda a,b: a-b, origin, orig)
            if offset[0] % st[0] or offset[1] % st[1] or offset[2] % st[2]:
                continue                # Offset stagger.
            moffset = map(lambda o,s: o / s, offset, st)
            msize = map(lambda s,t: (s+t-1) / t, size, st)
            m = matrix[moffset[2]:moffset[2]+msize[2]:dstep[2],
                                 moffset[1]:moffset[1]+msize[1]:dstep[1],
                                 moffset[0]:moffset[0]+msize[0]:dstep[0]]
            dcache.lookup_data(key) # update access time
            return m

        return None

    # ---------------------------------------------------------------------------
    #
    def cache_data(self, m, origin, size, step):

        dcache = self.data_cache
        if dcache is None:
            return

        key = (self, tuple(origin), tuple(size), tuple(step))
        elements = reduce(lambda a,b: a*b, m.shape, 1)
        bytes1 = elements * m.itemsize
        groups = [self]
        descrip = self.data_description(origin, size, step)
        dcache.cache_data(key, m, bytes1, descrip, groups)

    # ---------------------------------------------------------------------------
    #
    def data_description(self, origin, size, step):

        description = self.name

        if origin == (0,0,0):
            bounds = ' (%d,%d,%d)' % tuple(size)
        else:
            region = (origin[0], origin[0]+size[0]-1,
                                origin[1], origin[1]+size[1]-1,
                                origin[2], origin[2]+size[2]-1)
            bounds = ' (%d-%d,%d-%d,%d-%d)' % region
        description += bounds

        if step != (1,1,1):
            description += ' step (%d,%d,%d)' % tuple(step)

        return description

    # ---------------------------------------------------------------------------
    #
    def clear_cache(self):

        dcache = self.data_cache
        if dcache is None:
            return

        for k,d in dcache.group_keys_and_data(self):
            dcache.remove_key(k)

    # ---------------------------------------------------------------------------
    #
    def add_change_callback(self, cb):

        self.change_callbacks.append(cb)

    # ---------------------------------------------------------------------------
    #
    def remove_change_callback(self, cb):

        self.change_callbacks.remove(cb)

    # ---------------------------------------------------------------------------
    # Code has modified matrix elements, or the value type has changed.
    #
    def values_changed(self):

        self.call_callbacks('values changed')

    # ---------------------------------------------------------------------------
    # Mapping of array indices to xyz coordinates has changed.
    #
    def coordinates_changed(self):

        self.call_callbacks('coordinates changed')

    # ---------------------------------------------------------------------------
    #
    def call_callbacks(self, reason):

        for cb in self.change_callbacks:
            cb(reason)

###############################################################################

class MRC_Data:

    def __init__(self, path, file_type):


        self.path = path
        self.name = os.path.basename(path)

        file1 = open(path, 'rb')

        file1.seek(0,2)                                                            # go to end of file
        file_size = file1.tell()
        file1.seek(0,0)                                                          # go to beginning of file

        # Infer file byte order from column axis size nc.    Requires nc < 2**16
        # Was using mode value but 0 is allowed and does not determine byte order.
        self.swap_bytes = 0
        nc = self.read_values(file1, np.int32, 1)
        self.swap_bytes = not (nc > 0 and nc < 65536)
        file1.seek(0,0)

        v = self.read_header_values(file1, file_size, file_type)

        if v.get('imodStamp') == 1146047817:
            unsigned_8_bit = (v['imodFlags'] & 0x1 == 0)
        else:
            unsigned_8_bit = (file_type == 'imod' or v['type'] == 'mrc')
        self.element_type = self.value_type(v['mode'], unsigned_8_bit)

        self.check_header_values(v, file_size, file1)
        self.header = v                         # For dumpmrc.py standalone program.

        self.data_offset = file1.tell()
        file1.close()

        # Axes permutation.
        # Names c,r,s refer to fast, medium, slow file matrix axes.
        # Names i,j,k refer to x,y,z spatial axes.
        mapc, mapr, maps = v['mapc'], v['mapr'], v['maps']
        if (1 in (mapc, mapr, maps) and
                2 in (mapc, mapr, maps) and
                3 in (mapc, mapr, maps)):
            crs_to_ijk = (mapc-1,mapr-1,maps-1)
            ijk_to_crs = [None,None,None]
            for a in range(3):
                ijk_to_crs[crs_to_ijk[a]] = a
        else:
            crs_to_ijk = ijk_to_crs = (0, 1, 2)
        self.crs_to_ijk = crs_to_ijk
        self.ijk_to_crs = ijk_to_crs

        crs_size = v['nc'], v['nr'], v['ns']
        self.matrix_size = [int(s) for s in crs_size]
        self.data_size = [int(crs_size[a]) for a in ijk_to_crs]

        mx, my, mz = v['mx'], v['my'], v['mz']
        xlen, ylen, zlen = v['xlen'], v['ylen'], v['zlen']
        if mx > 0 and my > 0 and mz > 0 and xlen > 0 and ylen > 0 and zlen > 0:
            self.data_step = (xlen/mx, ylen/my, zlen/mz)
        else:
            self.data_step = (1.0, 1.0, 1.0)

        alpha, beta, gamma = (v['alpha'], v['beta'], v['gamma'])
        if not valid_cell_angles(alpha, beta, gamma, path):
            alpha = beta = gamma = 90
        self.cell_angles = (alpha, beta, gamma)

        if (v['type'] == 'mrc2000' and
                (v['zorigin'] != 0 or v['xorigin'] != 0 or v['yorigin'] != 0)):
            #
            # This is a new MRC 2000 format file.    The xyz origin header parameters
            # are used instead of using ncstart, nrstart nsstart for new style files,
            # provided the xyz origin specified is not zero.    It turns out the
            # xorigin, yorigin, zorigin values are zero in alot of new files while
            # the ncstart, nrstart, nsstart give the correct (non-zero) origin. So in
            # cases where the xyz origin parameters and older nrstart, ncstart,
            # nsstart parameters specify different origins the one that is non-zero
            # is preferred.    And if both are non-zero, the newer xorigin, yorigin,
            # zorigin are used.
            #
            self.data_origin = (v['xorigin'], v['yorigin'], v['zorigin'])
        else:
            crs_start = v['ncstart'], v['nrstart'], v['nsstart']
            ijk_start = [crs_start[a] for a in ijk_to_crs]
            # Check if ijk_start values appear to be uninitialized.
            limit = 10*max(max(mx,my,mz), max(self.data_size))
            if [s for s in ijk_start if abs(s) > limit]:
                self.data_origin = (0., 0., 0.)
            else:
                self.data_origin = scale_and_skew(ijk_start, self.data_step, self.cell_angles)

        r = ((1,0,0),(0,1,0),(0,0,1))
        for lbl in v['labels']:
            if lbl.startswith(b'Chimera rotation: '):
                ax,ay,az,angle = map(float, lbl.rstrip('\0').split()[2:])

                #S = Structure()
                r = self.rotation_matrix([ax, ay, az], angle)
                #r = Matrix.rotation_from_axis_angle((ax,ay,az), angle)

        self.rotation = r

        self.min_intensity = v['amin']
        self.max_intensity = v['amax']



    def rotation_matrix(self, axis, theta):
        '''
        compute matrix needed to rotate the system around an arbitrary axis (using Euler-Rodrigues formula).

        :param axis: 3d vector (numpy array), representing the axis around which to rotate
        :param theta: desired rotation angle
        :returns: 3x3 rotation matrix
        '''

        # if rotation angle is equal to zero, no rotation is needed
        if theta == 0:
            return np.identity(3)

        # method taken from
        # http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector
        axis = axis / np.sqrt(np.dot(axis, axis))
        a = np.cos(theta / 2)
        b, c, d = -axis * np.sin(theta / 2)
        return np.array([[a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)],
                         [2 * (b * c + a * d), a * a + c * c - b * b - d * d, 2 * (c * d - a * b)],
                         [2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c]])




    # Format derived from C header file mrc.h.
    #
    def read_header_values(self, file1, file_size, file_type):

        MRC_USER = 29
        CCP4_USER = 15
        MRC_NUM_LABELS = 10
        MRC_LABEL_SIZE = 80
        MRC_HEADER_LENGTH = 1024

        i32 = np.int32
        f32 = np.float32

        v = {}
        v['nc'], v['nr'], v['ns'] = self.read_values(file1, i32, 3)
        v['mode'] = self.read_values(file1, i32, 1)
        v['ncstart'], v['nrstart'], v['nsstart'] = self.read_values(file1, i32, 3)
        v['mx'], v['my'], v['mz'] = self.read_values(file1, i32, 3)
        v['xlen'], v['ylen'], v['zlen'] = self.read_values(file1, f32, 3)
        v['alpha'], v['beta'], v['gamma'] = self.read_values(file1, f32, 3)
        v['mapc'], v['mapr'], v['maps'] = self.read_values(file1, i32, 3)
        v['amin'], v['amax'], v['amean'] = self.read_values(file1, f32, 3)
        v['ispg'], v['nsymbt'] = self.read_values(file1, i32, 2)
        if file_type == 'ccp4':
            v['lskflg'] = self.read_values(file1, i32, 1)
            v['skwmat'] = self.read_values(file1, f32, 9)
            v['skwtrn'] = self.read_values(file1, f32, 3)
            v['user'] = self.read_values(file1, i32, CCP4_USER)
            v['map'] = file1.read(4)     # Should be 'MAP '.
            v['machst'] = self.read_values(file1, i32, 1)
            v['rms'] = self.read_values(file1, f32, 1)
            v['type'] = 'ccp4'
        else:
            # MRC file
            user = file1.read(4*MRC_USER)
            if user[-4:] == 'MAP ':
                # New style MRC 2000 format file with xyz origin
                v['user'] = self.read_values_from_string(user, i32, MRC_USER)[:-4]
                xyz_origin = self.read_values_from_string(user[-16:-4], f32, 3)
                v['xorigin'], v['yorigin'], v['zorigin'] = xyz_origin
                v['imodStamp'] = self.read_values_from_string(user[56:60], i32, 1)
                v['imodFlags'] = self.read_values_from_string(user[60:64], i32, 1)
                v['machst'] = self.read_values(file1, i32, 1)
                v['rms'] = self.read_values(file1, f32, 1)
                v['type'] = 'mrc2000'
            else:
                # Old style MRC has xy origin instead of machst and rms.
                v['user'] = self.read_values_from_string(user, i32, MRC_USER)
                v['xorigin'], v['yorigin'] = self.read_values(file1, f32, 2)
                v['type'] = 'mrc'

        v['nlabl'] = self.read_values(file1, i32, 1)
        labels = []
        for i in range(MRC_NUM_LABELS):
            labels.append(file1.read(MRC_LABEL_SIZE))
        v['labels'] = labels

        # Catch incorrect nsymbt value.
        if v['nsymbt'] < 0 or v['nsymbt'] + MRC_HEADER_LENGTH > file_size:
            raise SyntaxError('MRC header value nsymbt (%d) is invalid'
                                                    % v['nsymbt'])
        v['symop'] = file1.read(v['nsymbt'])

        return v

    #
    def value_type(self, mode, unsigned_8_bit):

        MODE_char     = 0
        MODE_short    = 1
        MODE_float    = 2

        if mode == MODE_char:
            if unsigned_8_bit:
                t = np.dtype(np.uint8)
            else:
                t = np.dtype(np.int8)                # CCP4 or MRC2000
        elif mode == MODE_short:
            t = np.dtype(np.int16)
        elif mode == MODE_float:
            t = np.dtype(np.float32)
        else:
            raise SyntaxError('MRC data value type (%d) ' % mode +
                                                    'is not 8 or 16 bit integers or 32 bit floats')

        return t

    #
    def check_header_values(self, v, file_size, file1):

        if v['nc'] <= 0 or v['nr'] <= 0 or v['ns'] <= 0:
            raise SyntaxError('Bad MRC grid size (%d,%d,%d)'
                                                    % (v['nc'],v['nr'],v['ns']))

        esize = self.element_type.itemsize
        data_size = int(v['nc']) * int(v['nr']) * int(v['ns']) * esize
        header_end = file1.tell()
        if header_end + data_size > file_size:
            if v['nsymbt'] and (header_end - v['nsymbt']) + data_size == file_size:
                # Sometimes header indicates symmetry operators are present but
                # they are not.    This error occurs in macromolecular structure database
                # entries emd_1042.map, emd_1048.map, emd_1089.map, ....
                # This work around code allows the incorrect files to be read.
                file1.seek(-v['nsymbt'], 1)
                v['symop'] = ''
            else:
                msg = ('File size %d too small for grid size (%d,%d,%d)'
                             % (file_size, v['nc'],v['nr'],v['ns']))
                if v['nsymbt']:
                    msg += ' and %d bytes of symmetry operators' % (v['nsymbt'],)
                raise SyntaxError(msg)

    #
    def read_values(self, file1, etype, count):

        esize = np.array((), etype).itemsize
        string = file1.read(esize * count)
        if len(string) < esize * count:
            raise SyntaxError('MRC file is truncated.    Failed reading %d values, type %s' % (count, etype.__name__))
        values = self.read_values_from_string(string, etype, count)
        return values

    #
    def read_values_from_string(self, string, etype, count):

        values = np.frombuffer(string, etype)
        if self.swap_bytes:
            values = values.byteswap()
        if count == 1:
            return values[0]
        return values

    # Reads a submatrix from a the file.
    # Returns 3d numpy matrix with zyx index order.
    #
    def read_matrix(self, ijk_origin, ijk_size, ijk_step, progress):

        # ijk correspond to xyz.    crs refers to fast,medium,slow matrix file axes.
        crs_origin = [ijk_origin[a] for a in self.crs_to_ijk]
        crs_size = [ijk_size[a] for a in self.crs_to_ijk]
        crs_step = [ijk_step[a] for a in self.crs_to_ijk]

        matrix = read_array(self.path, self.data_offset,
                                                crs_origin, crs_size, crs_step,
                                                self.matrix_size, self.element_type, self.swap_bytes,
                                                progress)
        if not matrix is None:
            matrix = self.permute_matrix_to_xyz_axis_order(matrix)

        return matrix

    #
    def permute_matrix_to_xyz_axis_order(self, matrix):

        if self.ijk_to_crs == (0,1,2):
            return matrix

        kji_to_src = [2-self.ijk_to_crs[2-a] for a in (0,1,2)]
        m = matrix.transpose(kji_to_src)

        return m
###############################################################################
###############################################################################

def valid_cell_angles(alpha, beta, gamma, path):

    err = None

    for a in (alpha, beta, gamma):
        if a <= 0 or a >= 180:
            err = 'must be between 0 and 180'

    if alpha + beta + gamma >= 360 and err is None:
        err = 'sum must be less than 360'

    if max((alpha, beta, gamma)) >= 0.5 * (alpha + beta + gamma) and err is None:
        err = 'largest angle must be less than sum of other two'

    if err:
        raise Exception('%s: invalid cell angles %.5g,%.5g,%.5g %s.\n'
                                 % (path, alpha, beta, gamma, err))
        return False

    return True

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Return 3 by 4 matrix where first 3 columns give rotation and last column
# is translation.
#
def transformation_and_inverse(origin, step, axes):

    ox, oy, oz = origin
    d0, d1, d2 = step
    ax, ay, az = axes

    tf = ((d0*ax[0], d1*ay[0], d2*az[0], ox),
                (d0*ax[1], d1*ay[1], d2*az[1], oy),
                (d0*ax[2], d1*ay[2], d2*az[2], oz))

    #from Matrix import invert_matrix
    #tf_inv = invert_matrix(tf)
    tf_inv=tf

    # Replace array by tuples
    tf_inv = tuple(map(tuple, tf_inv))

    return tf, tf_inv

# -----------------------------------------------------------------------------
# Apply scaling and skewing transformations.
#
def scale_and_skew(ijk, step, cell_angles):

    # Convert to radians
    alpha, beta, gamma = map(lambda a: a * np.pi / 180, cell_angles)

    cg = np.cos(gamma)
    sg = np.sin(gamma)
    cb = np.cos(beta)
    ca = np.cos(alpha)
    c1 = (ca - cb*cg)/sg
    c2 = np.sqrt(1 - cb*cb - c1*c1)

    i, j, k = ijk
    d0, d1, d2 = step

    xyz = (d0*i + d1*cg*j + d2*cb*k, d1*sg*j + d2*c1*k, d2*c2*k)
    return xyz


# -----------------------------------------------------------------------------
#
def apply_rotation(r, v):

    rv = [r[a][0]*v[0] + r[a][1]*v[1] + r[a][2]*v[2] for a in (0,1,2)]
    return tuple(rv)

# -----------------------------------------------------------------------------
#
def map_point(p, tf):

    tfp = [0,0,0]
    for r in range(3):
        tfr = tf[r]
        tfp[r] = tfr[0]*p[0] +tfr[1]*p[1] + tfr[2]*p[2] + tfr[3]
    tfp = tuple(tfp)
    return tfp

# -----------------------------------------------------------------------------

def read_array(path, byte_offset, ijk_origin, ijk_size, ijk_step,
                             full_size, type1, byte_swap, progress = None):

        if (tuple(ijk_origin) == (0,0,0) and
                tuple(ijk_size) == tuple(full_size) and
                tuple(ijk_step) == (1,1,1)):
                m = read_full_array(path, byte_offset, full_size,
                                                        type1, byte_swap, progress)
                return m

        matrix = allocate_array(ijk_size, type1, ijk_step, progress)

        file1 = open(path, 'rb')

        if progress:
                progress.close_on_cancel(file1)

        # Seek in file to read needed 1d slices.
        io, jo, ko = ijk_origin
        isize, jsize, ksize = ijk_size
        istep, jstep, kstep = ijk_step
        element_size = matrix.itemsize
        jbytes = full_size[0] * element_size
        kbytes = full_size[1] * jbytes
        ibytes = isize * element_size
        ioffset = io * element_size
        from numpy import frombuffer
        for k in range(ko, ko+ksize, kstep):
            if progress:
                progress.plane((k-ko)/kstep)
            kbase = byte_offset + k * kbytes
            for j in range(jo, jo+jsize, jstep):
                offset = kbase + j * jbytes + ioffset
                file1.seek(offset)
                data = file1.read(ibytes)
                slice1 = frombuffer(data, type1)
                matrix[int((k-ko)/kstep), int((j-jo)/jstep), :] = slice1[::istep]

        file1.close()

        if byte_swap:
            matrix.byteswap(True)

        return matrix

# -----------------------------------------------------------------------------
# Read an array from a binary file making at most one copy of array in memory.
#
def read_full_array(path, byte_offset, size, type1, byte_swap,
                                        progress = None, block_size = 2**20):

        a = allocate_array(size, type1)

        file1 = open(path, 'rb')
        file1.seek(byte_offset)

        if progress:
                progress.close_on_cancel(file1)
                a_1d = a.ravel()
                n = len(a_1d)
                nf = float(n)
                for s in range(0,n,block_size):
                        b = a_1d[s:s+block_size]
                        file1.readinto(b)
                        progress.fraction(s/nf)
                progress.done()
        else:
                file1.readinto(a)

        file1.close()

        if byte_swap:
                a.byteswap(True)

        return a

# -----------------------------------------------------------------------------
# Read ascii float values on as many lines as needed to get count values.
#
def read_text_floats(path, byte_offset, size, array = None,
                                         transpose = False, line_format = None, progress = None):

        if array is None:
                shape = list(size)
                if not transpose:
                        shape.reverse()
                from numpy import zeros, float32
                array = zeros(shape, float32)

        f = open(path, 'rb')

        if progress:
                f.seek(0,2)         # End of file
                file_size = f.tell()
                progress.text_file_size(file_size)
                progress.close_on_cancel(f)

        f.seek(byte_offset)

        try:
                read_float_lines(f, array, line_format, progress)
        except SyntaxError as msg:
                f.close()
                raise

        f.close()

        if transpose:
                array = array.transpose()

        if progress:
                progress.done()

        return array

# -----------------------------------------------------------------------------
#
def read_float_lines(f, array, line_format, progress = None):

        a_1d = array.ravel()
        count = len(a_1d)

        c = 0
        while c < count:
                line = f.readline()
                if line == '':
                        msg = ('Too few data values in %s, found %d, expecting %d'
                                     % (f.name, c, count))
                        raise SyntaxError(msg)
                if line[0] == '#':
                        continue                                    # Comment line
                if line_format is None:
                        fields = line.split()
                else:
                        fields = split_fields(line, *line_format)
                if c + len(fields) > count:
                        fields = fields[:count-c]
                try:
                        values = map(float, fields)
                except:
                        msg = 'Bad number format in %s, line\n%s' % (f.name, line)
                        raise SyntaxError(msg)
                for v in values:
                        a_1d[c] = v
                        c += 1
                if progress:
                        progress.fraction(float(c)/(count-1))

# -----------------------------------------------------------------------------
#
def split_fields(line, field_size, max_fields):

    fields = []
    for k in range(0, len(line), field_size):
        f = line[k:k+field_size].strip()
        if f:
            fields.append(f)
        else:
            break
    return fields[:max_fields]

# -----------------------------------------------------------------------------
#
def allocate_array(size, value_type = np.float32, step = None, progress = None,
                                     reverse_indices = True, zero_fill = False):




        if step is None:
                msize = size
        else:
                msize = [1+(sz-1)/st for sz,st in zip(size, step)]

        msize = np.array(msize).astype(int)

        shape = list(msize)
        if reverse_indices:
                shape.reverse()

        if zero_fill:
                from numpy import zeros as alloc
        else:
                from numpy import empty as alloc

        m = alloc(shape, value_type)
        return m

# -----------------------------------------------------------------------------
#
def closest_mrc2000_type(type1):

        if type1 in (np.float32, np.float64, np.int32, np.uint32, np.uint, np.uint16):
                ctype = np.float32
        elif type1 in (np.int16, np.uint8):
                ctype = np.int16
        elif type1 in (np.int8, np.int0, np.character):
                ctype = np.int8
        else:
                raise TypeError('Volume data has unrecognized type %s' % type1)

        return ctype

# -----------------------------------------------------------------------------
#can read 'mrc' or 'ccp4' or 'imod'
def read_density(filename,extension):

    try:
        grid_data= MRC_Grid(filename, extension)
    except Exception as e:
        raise Exception('cannot load density map %s: %s'%(filename, e))

    mtype = grid_data.value_type.type
    try:
        type1 = closest_mrc2000_type(mtype)
    except Exception as e:
        raise Exception("%s"%e)

    isz, jsz, ksz = grid_data.size
    m=[]
    for k in range(ksz):
        matrix = grid_data.matrix((0,0,k), (isz,jsz,1))
        if type1 != mtype:
            matrix = matrix.astype(type1)

        m.append(matrix)

    data=np.squeeze(np.array(m).astype(float))

    #return data, grid_data
    return np.swapaxes(data,0,2), grid_data


data_cache = Data_Cache(size = 0)

if __name__=="__main__":

    import os
    filename = "..%stest%sEMD-1080.mrc"%(os.sep, os.sep)

    #try:
    [density,data] = read_density(filename, 'mrc')
    print("origin: %s"%np.array(data.origin))
    print("shape: %s"%np.array(density.shape))
    print("delta: %s"%(np.identity(3)*np.array(data.mrc_data.data_step)))

    #except Exception as e:
    #    print("%s"%e)



# Copyright (c) 2014-2022 Matteo Degiacomi
#
# BiobOx is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# BiobOx is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with BiobOx ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
#
# Author : Matteo Degiacomi, matteo.degiacomi@gmail.com

import os

import scipy.ndimage.filters
from sklearn.cluster import DBSCAN
import numpy as np
import pandas as pd

from biobox.classes.structure import Structure

class Density(Structure):
    '''
    Subclass of :func:`Structure <structure.Structure>`, allows importing density map, and transform them in a PDB file containing a collection of spheres placed on the map's high density regions.
    '''

    def __init__(self):
        '''
        A density map is fully described by the following attributes, stored in the self.properties dictionary:

        :param density: density map
        :param delta:   scaling factor for voxels (default is [1, 1, 1] Angstrom)
        :param size:    dimensions in voxels
        :param origin:  bottom-left-front corner of the cube
        :param radius: radius of points composing the density map
        :param format: format name (only dx supported at the moment)
        '''

        super(Density, self).__init__()
        self._reset_info()

    def _reset_info(self, r=1.9):
        '''
        reset all properties related to a density map (clean)
        '''

        self.properties['density'] = np.array([])  # density map

        self.properties['radius'] = r
        # scaling factor for voxels (default is [1, 1, 1] Angstrom)
        self.properties['delta'] = np.array([])
        self.properties['size'] = np.array([])  # dimensions in voxels
        # bottom-left-front corner of the cube
        self.properties['origin'] = np.array([])
        self.properties['format'] = ""  # file format
        self.properties['scan'] = np.array([])
        self.properties['filename'] = ""

        self.clear()

    def return_density_map(self):
        '''
        :returns: density map as 3D numpy array
        '''
        return self.properties['density']

    def import_map(self, filename, fileformat='dx'):
        '''
        Import density map and fill up the points and properties data structures.

        :param  filename: name of density file to load
        :param  fileformat: at the moment supports dx, ccp4, mrc and imod
        '''

        if not os.path.exists(filename):
            raise Exception("%s not found!" % filename)

        # call format-specific loading functions.
        # function should fill up all required properties in _reset_info(), and
        # load the map as a 3D array, containing intensity values.
        try:
            if fileformat == 'dx':
                self._import_dx(filename)
            elif fileformat == 'ccp4':
                self._import_mrc(filename, 'ccp4')
            elif fileformat == 'mrc':
                self._import_mrc(filename, 'mrc')
            elif fileformat == 'imod':
                self._import_mrc(filename, 'imod')
            else:
                raise Exception("sorry, format %s is not supported" % fileformat)

        except Exception as e:
            self._reset_info()
            Exception("ERROR: %s" % e)

        # if any error went undetected during loading (missing information),
        # data structures may be inconsistent. Call cleaning procedure!
        if len(self.properties['density']) == 0:
            print("density map could not be correctly loaded!")
            self._reset_info(self)
        elif len(self.properties['size']) == 0:
            print("density map information missing!")
            self._reset_info(self)
        elif len(self.properties['origin']) == 0:
            print("map origin information missing!")
            self._reset_info(self)
        elif len(self.properties['delta']) == 0:
            print("voxel size information missing!")
            self._reset_info(self)

        # if all required information is present, place points instead of
        # voxels
        else:
            try:
                self.place_points()
            except Exception as e:
                pass

            self.properties['format'] = format
            self.properties['filename'] = filename

        self.properties["sigma"] = np.std(self.properties['density'])

    def import_numpy(self, data, origin=[0, 0, 0], delta=np.identity(3)):
        '''
        import a numpy 3D array to allow manipulation as a density map

        :param data: numpy 3D array
        :param origin: coordinates of bottom left corner of the map
        :param delta: voxels' shape (default is a cubic voxel of 1 Angstrom-long sides).
        '''

        if len(data.shape) != 3:
            raise Exception("ERROR: a 3D numpy array is expected")

        self.properties['density'] = data
        self.properties['origin'] = np.array(origin)
        self.properties['size'] = np.array(data.shape)
        self.properties['delta'] = delta

        # sphere size corresponding to the volume of one voxel
        voxel_volume = self.properties['delta'][0, 0] * self.properties['delta'][1, 1] * self.properties['delta'][2, 2]
        self.properties['radius'] = (voxel_volume * 3 / (4 * np.pi))**(1 / 3.0)


    def get_oversampled_points(self, sigma=0):
        '''
        return points obtained by oversampling the map (doule points on every axis)

        :param sigma: place points only on voxels having intensity greater than threshold
        :returns: points 3D points placed on voxels having value higher than threshold
        :returns: radius radius of produced points
       '''

        thresh = self.get_thresh_from_sigma(sigma)

        oversampled_data = np.zeros((self.properties['size'][0] * 2,
                                     self.properties['size'][1] * 2,
                                     self.properties['size'][2] * 2))

        for x in range(1, self.properties['size'][0] - 1, 1):
            for y in range(1, self.properties['size'][1] - 1, 1):
                for z in range(1, self.properties['size'][2] - 1, 1):
                    if self.properties['density'][x, y, z] > thresh:
                        oversampled_data[x * 2, y * 2, z * 2] = self.properties['density'][x, y, z]
                        oversampled_data[x * 2 + 1, y * 2, z * 2] = self.properties['density'][x, y, z]
                        oversampled_data[x * 2, y * 2 + 1, z * 2] = self.properties['density'][x, y, z]
                        oversampled_data[x * 2, y * 2, z * 2 + 1] = self.properties['density'][x, y, z]
                        oversampled_data[x * 2 - 1, y * 2, z * 2] = self.properties['density'][x, y, z]
                        oversampled_data[x * 2, y * 2 - 1, z * 2] = self.properties['density'][x, y, z]
                        oversampled_data[x * 2, y * 2, z * 2 - 1] = self.properties['density'][x, y, z]

        delta = self.properties['delta'] / 2
        radius = self.properties['radius'] * 2 / 3

        # define scaling to shrink everything by a size equal to spheres radius
        size = np.max(np.transpose(np.where(oversampled_data > thresh)), axis=0) - np.min(np.transpose(np.where(oversampled_data > thresh)), axis=0)
        scaled_size = np.max(np.transpose(np.where(oversampled_data > thresh)) * np.diag(delta), axis=0) - np.min(np.transpose(np.where(oversampled_data > thresh)) * np.diag(delta), axis=0)

        scaling = 8.0 / 3.0
        scale_x = (delta[0, 0] - 1.0) / (scaled_size[0] - size[0]) * (scaled_size[0] - radius * scaling)
        scale_y = (delta[1, 1] - 1.0) / (scaled_size[1] - size[1]) * (scaled_size[1] - radius * scaling)
        scale_z = (delta[2, 2] - 1.0) / (scaled_size[2] - size[2]) * (scaled_size[2] - radius * scaling)

        # create structure data (ensemble of points and their center, and store
        # its geometric center)
        points = np.transpose(np.where(oversampled_data > thresh)) * np.array([scale_x, scale_y, scale_z]) + self.properties['origin'] + np.ones(3) * self.properties['radius'] * 2 / 3
        return points, radius

    def get_thresh_from_sigma(self, val):
        '''
        convert cutoff value from sigma multiples into actual threshold

        :param val: sigma scaling
        :returns: cutoff
        '''

        return self.properties["sigma"] * val

    def get_sigma_from_thresh(self, t):
        '''
        convert cutoff value from actual threshold to sigma multiple

        :param threshold value
        :returns sigma multiple
        '''

        return t / float(self.properties["sigma"])

    def place_points(self, sigma=0, noise_filter=0.01):
        '''
        given density information, place points on every voxel above a given threshold.

        :param sigma: intensity threshold value.
        :param noise_filter: launch DBSCAN clustering algorithm to detect connected regions in density map. Regions representing less than noise_filter of the total will be removed. This is a ratio, value should be between 0 and 1.
        '''

        thresh = self.get_thresh_from_sigma(sigma)

        if not np.any(self.properties['density'] > thresh):
            raise IOError("selected threshold leads to empty point ensemble")

        if noise_filter >= 1 or noise_filter < 0:
            raise IOError("noise_filter should be between 0 and 1")

        # define scaling to shrink everything by a size equal to spheres radius
        size = np.max(np.transpose(np.where(self.properties['density'] > thresh)), axis=0) - np.min(np.transpose(np.where(self.properties['density'] > thresh)), axis=0)
        scaled_size = np.max(np.transpose(np.where(self.properties['density'] > thresh)) * np.diag(self.properties['delta']), axis=0)-np.min(np.transpose(np.where(self.properties['density'] > thresh)) * np.diag(self.properties['delta']), axis=0)

        if self.properties['delta'][0, 0] != 1:
            scale_x = (self.properties['delta'][0, 0] - 1.0) / (scaled_size[0] - size[0]) * (scaled_size[0] - self.properties['radius'])
        else:
            scale_x = 1.0

        if self.properties['delta'][1, 1] != 1:
            scale_y = (self.properties['delta'][1, 1] - 1.0) / (scaled_size[1] - size[1]) * (scaled_size[1] - self.properties['radius'])
        else:
            scale_y = 1.0

        if self.properties['delta'][2, 2] != 1:
            scale_z = (self.properties['delta'][2, 2] - 1.0) / (scaled_size[2] - size[2]) * (scaled_size[2] - self.properties['radius'])
        else:
            scale_z = 1
        # create structure data (ensemble of points and their center, and store
        # its geometric center)
        points = np.transpose(np.where(self.properties['density'] > thresh)) * np.array([scale_x, scale_y, scale_z]) + self.properties['origin'] + np.ones(3) * self.properties['radius']

        # remove previous points arrangement, and create new one (necessary,
        # since the amount of points will change, and cannot therefore be
        # considered a new conformation)
        self.clear()

        # apply DBSCAN noise filter
        if noise_filter != 0:
            step = self.properties['delta'][0, 0] * np.sqrt(3)
            db = DBSCAN(eps=step, min_samples=10).fit(points)
            pts2 = []
            for i in np.unique(db.labels_):
                if np.sum(db.labels_ == i) / float(len(points)) > 0.01 and i != -1:
                    if len(pts2) == 0:
                        pts2 = points[db.labels_ == i]
                    else:
                        pts2 = np.concatenate((pts2, points[db.labels_ == i]))

            self.add_xyz(pts2)

        else:
            self.add_xyz(points)

        # update radii list (useful for instance for CCS calculation)
        idx = np.arange(len(self.points))
        self.data = pd.DataFrame(self.properties["radius"], index=idx, columns=["radius"])

        # self.get_center()

    def predict_ccs_from_mass(self, resolution, mass, density=0.84, x0=2.51911893, y0=-1.06481492, c=2.7018764, k=0.44084821):

        '''
        given target mass and map resolution, predict CCS. Mass threshold is rescaled using the fitting function c / (1 + exp(-k*(resolution-x0))) + y0.

        :param resolution: map resolution in 1/Angstrom
        :param mass: protein mass in kDa
        :param density: protein density in Da/A3
        :param x0: sigmoid parameter
        :param y0: sigmoid parameter
        :param c: sigmoid parameter
        :param k: sigmoid parameter
        :returns CCS estimated from mass and density map resolution
        '''

        if 'scan' not in list(self.properties):
            raise IOError("no threshold to volume to CCS relationship loaded yet. Please execute thresh_vol_ccs method.")

        data=self.properties['scan'].copy()
        data[:,1]*=density
        data[:,1]/=1000.0

        #get mass threshold
        dtest1=np.argmin(np.abs(data[:,1]-mass))
        thresh=data[dtest1,0]

        #rescale mass threshold
        scaling= c / (1 + np.exp(-k*(resolution-x0))) + y0
        calibrated_thresh=thresh/scaling

        #get associated CCS
        dtest1=np.argmin(np.abs(data[:,0]-calibrated_thresh))
        return data[dtest1,2], data[dtest1,0]



    def predict_mass_from_ccs(self, resolution, ccs, density=0.84, x0=2.51911893, y0=-1.06481492, c=2.7018764, k=0.44084821):
        '''
        given target mass and map resolution, predict CCS. Mass threshold is rescaled using the fitting function c / (1 + exp(-k*(resolution-x0))) + y0.

        :param resolution: map resolution in 1/Angstrom
        :param mass: protein mass in kDa
        :param density: protein density in Da/A3
        :param x0: sigmoid parameter
        :param y0: sigmoid parameter
        :param c: sigmoid parameter
        :param k: sigmoid parameter
        :returns CCS estimated from mass and density map resolution
        '''

        if 'scan' not in list(self.properties):
            raise IOError("no threshold to volume to CCS relationship loaded yet. Please execute thresh_vol_ccs method.")

        data=self.properties['scan'].copy()
        data[:,1]*=density
        data[:,1]/=1000.0

        #get mass threshold
        dtest1=np.argmin(np.abs(data[:,2]-ccs))
        thresh=data[dtest1,0]

        #rescale mass threshold
        scaling= c / (1 + np.exp(-k*(resolution-x0))) + y0
        calibrated_thresh=thresh*scaling

        #get associated CCS
        dtest1=np.argmin(np.abs(data[:,0]-calibrated_thresh))
        return data[dtest1,1], data[dtest1,0]


    def scan_threshold(self, mass, density=0.782878356, sampling_points=1000):
        '''
        if mass and density of object are known, filter the map on a linear scale of threshold values, and compare the obtained mass to the experimental one.

        .. note:: in proteins, an average value of 1.3 g/cm^3 (0.782878356 Da/A^3) can be assumed. Alternatively, the relation density=1.410+0.145*exp(-mass(kDa)/13) can be used.

        .. note:: 1 Da/A^3=0.602214120 g/cm^3

        :param mass: target mass in Da
        :param density: target density in Da/A^3
        :param sampling_points: number of measures to perform between min and max intensity in density map
        :returns: array reporting tested values and error on mass ([threshold, model_mass-target_mass])
        '''

        low = self.get_sigma_from_thresh(np.min(self.properties['density']))
        high = self.get_sigma_from_thresh(np.max(self.properties['density']))

        result = []
        for thresh in np.linspace(low, high, num=sampling_points):

            try:
                self.place_points(thresh)
                vol = self.get_volume()
            except Exception as e:
                vol = 0.0

            print("threshold=%s, error=%s" % (thresh, vol * density - mass))
            result.append([thresh, vol * density - mass])

        r = np.array(result)
        bestthresh = r[np.argmin(np.abs(r[:, 1]) - mass), 0]

        try:
            self.place_points(bestthresh)
        except Exception as ex:
            pass

        return r

    def find_data_from_sigma(self, sigma, exact=True, append=False, noise_filter=0.01):
        '''
        map experimental data to given threshold

        :param sigma: density threshold
        :param noise_filter: launch DBSCAN clustering algorithm to detect connected regions in density map. Regions representing less than noise_filter of the total will be removed. This is a ratio, value should be between 0 and 1.        '''

        thresh = self.get_sigma_from_thresh(sigma)

        if exact:
            import biobox as bb

            try:
                self.place_points(thresh, noise_filter)
                vol = self.get_volume()
                ccs = bb.ccs(self)
            except Exception as ex:
                vol = 0
                ccs = 0

            res = np.array([thresh, vol, ccs])

            if append:
                self.properties['scan'] = np.concatenate((self.properties['scan'], res))

            return res

        else:
            return self.properties['scan'][
                np.argmin(np.abs(self.properties['scan'][:, 0] - sigma))]


    def find_data_from_volume(self, vol):
        '''
        map experimental data to given volume

        :param vol: volume
        '''

        return self.properties['scan'][
            np.argmin(np.abs(self.properties['scan'][:, 1] - vol))]

    def find_data_from_ccs(self, ccs):
        '''
        map experimental data to given ccs

        :param ccs: target CCS (in A^2)
        '''

        return self.properties['scan'][
            np.argmin(np.abs(self.properties['scan'][:, 2] - ccs))]

    def threshold_vol_ccs(self, low="", high="", sampling_points=1000, append=False, noise_filter=0.01, verbose=False):
        '''
        return the volume to threshold to CCS relationship

        :param sampling_points: number of measures to perform between min and max intensity in density map
        :returns: array reporting tested values and error on mass ([threshold, model_mass-target_mass])
        '''

        import biobox as bb

        if low == "":
            low = self.get_sigma_from_thresh(np.min(self.properties['density']))
        if high == "":
            high = self.get_sigma_from_thresh(np.max(self.properties['density']))

        result = []
        for thresh in np.linspace(low, high, num=sampling_points):
            try:
                self.place_points(thresh, noise_filter=noise_filter)
                vol = self.get_volume()
                ccs = bb.ccs(self)
            except Exception as ex:
                vol = 0
                ccs = 0

            if verbose:
                print("thresh: %s, vol=%s, ccs=%s (%s points)" % (thresh, vol, ccs, len(self.points)))

            result.append([thresh, vol, ccs])

        r = np.array(result)

        if append:
            self.properties['scan'] = np.concatenate((self.properties['scan'], r))

        else:
            self.properties['scan'] = r

        return r

    def best_threshold(self, mass, density=0.782878356):
        '''
        If mass and density of object are known, try to filter the map so that the mass is best matched.

        search for best threshold using bissection method.

        .. note:: in proteins, an average value of 1.3 g/cm^3 (0.782878356 Da/A^3) can be assumed. Alternatively, the relation density=1.410+0.145*exp(-mass(kDa)/13) can be used.

        .. note:: 1 Da/A^3=1.660538946 g/cm^3
        .. note:: 1 g/cm^3=0.602214120 Da/A^3

        :param mass: target mass in Da
        :param density: target density in Da/A^3
        :returns: array reporting tested values and error on mass ([sigma, model_mass-target_mass])
        '''

        high = self.get_sigma_from_thresh(np.max(self.properties['density']))
        low = self.get_sigma_from_thresh(np.min(self.properties['density']))

        high_val = 1000000000
        low_val = -1000000000
        error = high_val

        result = []
        while True:
            thresh = (high + low) / 2.0

            try:
                self.place_points(thresh)
                vol = self.get_volume()
            except Exception as ex:
                vol = 0

            mass_model = vol * density
            error = mass_model - mass
            result.append([thresh, error])

            if error == high_val or error == low_val:
                if np.abs(high_val) < np.abs(low_val):
                    thresh = high
                else:
                    thresh = low
                break

            if error < 0:
                high = thresh
                high_val = error
            elif error > 0:
                low = thresh
                low_val = error
            else:
                break

        r = np.array(result)
        bestthresh = r[-1, 0]

        try:
            self.place_points(bestthresh)
        except Exception as ex:
            pass

        return r

    def blur(self, dimension=5, sigma=0.5):
        '''
        blur density applying a cubic gaussian kernel of given kernel dimension (cubic grid size).

        .. warning:: cannot be undone

        :param dimension: size of the kernel grid.
        :param sigma: standard deviation of gaussian kernel.
        '''

        shape = (dimension, dimension, dimension)

        m, n, k = [(ss - 1.) / 2. for ss in shape]

        x_ = np.arange(-m, m + 1, 1).astype(int)
        y_ = np.arange(-n, n + 1, 1).astype(int)
        z_ = np.arange(-k, k + 1, 1).astype(int)
        x, y, z = np.meshgrid(x_, y_, z_)

        h = np.exp(-(x * x + y * y) / (2. * sigma * sigma))
        h[h < np.finfo(h.dtype).eps * h.max()] = 0
        sumh = h.sum()
        if sumh != 0:
            h /= sumh

        dens = scipy.ndimage.filters.convolve(self.properties['density'], h, mode='constant')
        self.properties['density'] = dens
        self.properties["sigma"] = np.std(self.properties['density'])

    def write_dx(self, fname='dens.dx'):
        '''
        Write a density map in DX format

        :param fname: output file name
        '''

        dens = self.properties['density']
        origin = self.properties['origin']
        delta = self.properties['delta']

        fout = open(fname, "w")
        fout.write("# density generated with SBT\n#\n#\n#\n")
        fout.write("object 1 class gridpositions counts %s %s %s\n" % (dens.shape[0], dens.shape[1], dens.shape[2]))
        fout.write("origin %s %s %s\n"%(origin[0], origin[1], origin[2]))
        fout.write("delta %s %s %s\n"%(delta[0, 0], delta[0, 1], delta[0, 2]))
        fout.write("delta %s %s %s\n" % (delta[1, 0], delta[1, 1], delta[1, 2]))
        fout.write("delta %s %s %s\n" %(delta[2, 0], delta[2, 1], delta[2, 2]))
        fout.write("object 2 class gridconnections counts %s %s %s\n"%(dens.shape[0], dens.shape[1], dens.shape[2]))
        fout.write("object 3 class array type double rank 0 items %i data follows\n"%(dens.shape[0] * dens.shape[1] * dens.shape[2]))

        cnt = 0
        for xpos in range(0, dens.shape[0], 1):
            for ypos in range(0, dens.shape[1], 1):
                for zpos in range(0, dens.shape[2], 1):
                    fout.write("%s " % dens[xpos, ypos, zpos])
                    cnt += 1
                    if cnt % 3 == 0:
                        fout.write("\n")

        fout.close()

    def export_as_pdb(self, outname, step, threshold=0.1):
        '''
        Write a pdb file with points where the density exceeds a threshold

        :param outname: output file name
        :param step: stepsize used to generate the density map
        :param threshold: density to be exceeded to generate a point in pdb
        '''

        # @todo assign spheres beta factor to associated density value

        dens = self.properties['density']
        origin = self.properties['origin']

        fout = open(outname, 'w')

        cnt = 1
        identifier = 'ATOM'  # atom to be used to mimick density
        symbol = 'H'  # element for atom

        print('exporting density greater than %s to pdb' % threshold)

        for xpos in range(0, dens.shape[0], 1):
            for ypos in range(0, dens.shape[1], 1):
                for zpos in range(0, dens.shape[2], 1):
                    if dens[xpos, ypos, zpos] > threshold:
                        x_coord = float(xpos / step) + origin[0]
                        y_coord = float(ypos / step) + origin[1]
                        z_coord = float(zpos / step) + origin[2]
                        L = '%-6s%5s  %-4s%-4s  DIS    %8.3f%8.3f%8.3f  1.00  0.00            \n'%(identifier, cnt, symbol, symbol, x_coord, y_coord, z_coord)
                        fout.write(L)
                        cnt += 1

        fout.close()

    def _import_dx(self, filename):
        '''
        import density map and fill up the points and properties data structures.

        :param filename: name of dx file to load
        '''

        try:
            fin = open(filename, "r")
        except Exception as e:
            raise Exception('opening of file %s failed!' % filename)

        d = []
        dlt = []
        for line in fin:
            w = line.split()
            if w[0] == "#":
                continue
            elif len(w) <= 3: #and np.array(list(w)).dtype == ('float'):
                try:
                    w = np.array(w).astype(float)
                    for i in range(len(w)):
                        d.append(w[i])
                except Exception as e:
                    continue
                # get coordinates
            elif len(w) > 2 and w[0] == "object" and w[3] == "gridpositions":
                self.properties['size'] = np.array([w[-3], w[-2], w[-1]]).astype(int)
            # get scaling factor
            elif len(w) > 2 and w[0] == "delta":
                dlt.append(w[1:4])
            # get position of origin
            elif len(w) > 2 and w[0] == "origin":
                self.properties['origin'] = np.array(w[1:4]).astype(float)


        # scaling factor with respect of unit cell voxels
        self.properties['delta'] = np.array(dlt).astype(float)

        try:
            self.properties['density'] = np.reshape(
                np.array(d).astype(float),
                (self.properties['size'][0],
                 self.properties['size'][1],
                 self.properties['size'][2]))
        except Exception as ex:
            raise Exception("reshaping of dx data failed! Dimensions and dataset size are inconsistent!")

    def _import_mrc(self, filename, fileformat):
        '''
        import density map in MRC, CCP4 or IMOD format.

        MRC format here: www2.mrc-lmb.cam.ac.uk/image2000.html

        CCP4 format here: www.ccp4.ac.uk/html/maplib.html

        :param filename name of MRC or CCP4 file to load
        :param fileformat: can be mrc, imod or ccp4
        '''

        import biobox.classes.density_MRC as MRC
        try:
            [density, data] = MRC.read_density(filename, fileformat)
            self.properties['density'] = density
            self.properties['origin'] = np.array(data.origin)
            self.properties['size'] = np.array(density.shape)
            self.properties['delta'] = np.identity(3) * data.mrc_data.data_step

            # sphere size corresponding to the volume of one voxel
            voxel_volume = self.properties['delta'][0, 0] * self.properties['delta'][1, 1] * self.properties['delta'][2, 2]
            self.properties['radius'] = (voxel_volume * 3 / (4 * np.pi))**(1 / 3.0)

        except Exception as e:
            raise Exception("%s" % e)


    def get_volume(self):
        '''
        compute density map volume. This is done by counting the points, and multiplying that by voxels' volume.

        .. warning:: can be called only after :func:`place_points <density.Density.place_points>` has been called.

        .. warning:: supposes unskewed voxels.
        '''
        return self.properties['delta'][0, 0] * self.properties['delta'][1, 1] * self.properties['delta'][2, 2] * len(self.points)



if __name__ == "__main__":

    import biobox as bb

    print("loading density...")
    D = bb.Density()
    D.import_map("..%stest%sEMD-1080.mrc"%(os.sep, os.sep), "mrc")

    print("origin: %s"%np.array(D.properties["origin"]))
    print("shape: %s"%np.array(D.properties["density"].shape))
    print("delta: %s"%(D.properties["delta"]))

    # test points placement
    D.place_points(4)

    # test CCS calculation and prediction
    D.threshold_vol_ccs(sampling_points=50, append=False)#, noise_filter=1)
    ccs = D.predict_ccs_from_mass(11.5, 801)
    print(ccs)



# Copyright (c) 2014-2022 Matteo Degiacomi
#
# Biobox is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# Biobox is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with Biobox ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
#
# Author : Matteo Degiacomi, matteo.degiacomi@gmail.com

import os
from copy import deepcopy
import numpy as np
from biobox.classes.structure import Structure
from biobox.classes.assembly import Assembly


class Polyhedron(Assembly):
    '''
    Subclass of :func:`Assembly <assembly.Assembly>`, allowing the assembly of polyhedral symmetries.
    After instantiation, the first method to be called is :func:`setup_polyhedron <polyhedron.Polyhedron.setup_polyhedron>`
    '''

    def setup_polyhedron(self, polyname, M, dbfilename=""):
        '''
        load information for the generation of a polyhedral assembly.

        It loads the desired :func:`Structure <structure.Structure>` to be used as building block (it automatically centers and alignes it).
        it also loads and geometric information about the desired polyhedron from database.
        database contains this information: [polymer_name number_of_vertices numbe_of_edges vertices_coordinates vertices_connectivity connectivity_type]

        * Vertices_coordinates contains a list of 3D positions, formatted as x1 y1 z1 x2 y2 z2,...
        * Connectivity lists couples of vertices ID being connected, formatted as conn1_1 conn1_2 conn2_1 conn2_2...
        * conn_type allows to group edges, so that rotation angles can be applied to subgroups of those. This is a list of numbers, that should include zero as lowest number.

        A default conn_type is provided by database, the user can however regroup the edges at will.

        .. note:: This method has to be called first.

        :param polyname: name of the polyhedron to be assembled. The name should be located in the polyhedra database.
        :param M: monomer to be used as building block. Should be an instance of Structure class.
        :param dbfilename: polyhedra database file.
        '''

        if len(dbfilename) == 0:
            #"polyhedron_database.dat"
            folder = os.path.dirname(os.path.realpath(__file__))
            dbfilename = "%s/polyhedron_database.dat" % folder

        if os.path.isfile(dbfilename) != 1:
            raise Exception("ERROR: %s not found!" % dbfilename)

        # name of file containing polyhedra database
        self.dbfilename = dbfilename
        # name of polyhedron
        self.polyname = polyname

        # search information in new format database about desired polimer
        try:
            self.edges, self.v, self.conn, self.conn_type = self._search_database(polyname, dbfilename=self.dbfilename)
        except Exception as e:
            raise Exception("%s" % e)

        # monomeric subunit for polyhedron construction (instance of class
        # Molecule)
        self.building_block = deepcopy(M)
        self.building_block.align_axes()

        # box size
        dimensions = self.building_block.get_size()
        # box size along major axis (length)
        self.L = dimensions[0]
        # box size along second axis (width)
        self.W = dimensions[1]
        # box size along third axis (height)
        self.H = dimensions[2]

        # deformation database: [edge_id class x y z]. Edges having the same class will be scaled by the same coefficient.
        #@todo insert information about deformations directly into database. At the moment, it has to be explicitely provided via add_deformation method.
        self.deform = np.array([])

    def add_deformation(self, edges, vector=[]):
        '''
        Add entry in deformation database. Can be called only after the setup_polyhedron method.

        :param edges: integer or list of integers, defining edges indices that should be subjected to deformation.
        :param vector: axis along which deformation should take place. If not given, radial deformation will be assumed.
        '''

        # determine index for new entry in database
        if len(self.deform) == 0:
            idx = 0
        else:
            idx = np.max(self.deform[:, 1]) + 1

        # transform integer into list, in case an interger only is provided
        if isinstance(edges, int):
            e = [edges]
            edges = e

        # iterate over all edges and create new entries for database
        tmp = []
        for i in edges:

            # add deformation axis, if given
            if len(vector) == 3:
                vector /= np.linalg.norm(vector)
                tmp.append([i, idx, vector[0], vector[1], vector[2]])

            # axis not given, assume radial deformation
            else:
                ax = self.v[i] / np.linalg.norm(self.v[i])
                tmp.append([i, idx, ax[0], ax[1], ax[2]])

        # update deformation database
        if len(self.deform) > 0:
            self.deform = np.concatenate((self.deform, np.array(tmp)))
        else:
            self.deform = np.array(tmp)

    def generate_polyhedron(self, S, alpha, beta, gamma, deformation=[], add_conformation=False):
        '''
        Given polyhedron information previously loaded (using setup_polyhedron), generate an appropriate symmetry figure.

        :param S: polyhedron size (radius of mean sphere).
        :param alpha: rotation on molecule's first principal axis.
        :param beta: rotation on molecule's second principal axis.
        :param gamma: rotation on molecule's third principal axis.
        :param deformation: if list is provided, deformations will be applied according to self.deformation database.\n
               deformation list length should be equal to the amount of deformation classes)
        :param add_conformation: if True, the coordinates of the new poyhedron will be added to the conformational database as a new alternative conformation.\n
                If False, old polyhedron coordianates will be substituted.
        '''

        # if deformation coefficients are given, check first that they match
        # the number of classes in deformation database
        if len(deformation) > 0 and len(deformation) != len(np.unique(self.deform[:, 1])):
            raise Exception("ERROR: %s deformation coefficients expected, but %s found" % (len(deformation), np.unique(self.deform[:, 1])))

        self.psi, self.phi, self.nu, self.circumradius, self.midradius = self.get_polyhedron_properties(S)

        if not add_conformation:
            # clear any previous polyhedron and load anew
            self.clear()
            self.load(self.building_block, self.edges)
            self.current = 0
        else:
            # add structure to existing units as alternative conformation
            # (initially all identical)
            self.current += 1
            for i in range(0, len(self.unit), 1):
                self.unit[i].add_xyz(self.building_block.get_xyz())
                self.unit[i].set_current(self.current)

        # test type consistency: if arrays are provided, length should be equal
        # to the amount of different types in conn_type
        if 'ndarray' in str(type(alpha)):
            try:
                if len(alpha) != len(beta) or len(alpha) != len(gamma) or len(alpha) != len(np.unique(self.conn_type)):
                    print("ERROR: inconsistent length in provided angle arrays")
                    print("> received the following angles: %s, %s, %s" % (alpha, beta, gamma))
                    return -1
            except Exception as e:
                raise Exception("ERROR: all angle arrays should have length %s" %(len(np.unique(self.conn_type))))

        # generate desired polyhedral coordinates (note: internally we work in
        # radians, not degrees)
        poly_xyz = self._polycalc_core(self.W, self.L, self.H, S, self.nu, self.phi, np.radians(alpha), np.radians(beta), np.radians(gamma), deformation)

        for x in range(0, len(self.unit), 1):
            coords = poly_xyz[x, :, :].squeeze()

            # if add_conformation:
            #    self.unit[x].add_xyz(coords)
            # else:
            self.unit[x].set_xyz(coords)

    def set_current(self, index):
        '''
        Select current polyhedral conformation from ensemble of previously generated alternatives.
        This places the frame pointer at the same desired position in all polyhedron subunits.

        :param index: number of alternative conformation (starting from 0)
        '''
        for x in range(0, len(self.unit), 1):
            self.unit[x].set_current(index)

    def delete_xyz(self, index):
        '''
        Delete one conformation in the conformational database.
        the new current conformation will be the previous one

        :param index: alternative coordinates set to remove (starting from zero)
        '''

        for u in range(0, len(self.unit), 1):
            self.unit[u].delete_xyz(index)

        if index > 0:
            self.current = index - 1
        else:
            self.current = 0

    def rmsd_distance_matrix(self, points_indices=[]):
        '''
        Calculate the RMSD between all structures with respect of a reference structure.
        uses Kabsch alignement algorithm.

        :param points_indices: indices of points of interest. This must be a list of indices of atoms in unites, i.e. [[unit1_indices],[unit2_indices],...]
        :returns: RMSD of all structures with respect of reference structure (in a numpy array)
        '''

        # this method exploits the RMSD method implemented in Structure class.
        S = Structure()


        # all alternative coordinates are accumulated in a Structure instance
        for i in range(0, self.unit[0].coordinates.shape[0], 1):
            self.set_current(i)

            # if specific coordinates are requested, load only those
            if len(points_indices) > 0:
                pts = self.unit[0].get_xyz()
                for i in range(1, len(self.unit), 1):
                    pts = np.concatenate((pts, self.unit[i].get_xyz()[points_indices[i], :]))

            # otherwise, load everything
            else:
                pts = self.get_all_xyz()

            S.add_xyz(pts)

        return S.rmsd_distance_matrix()

    def write_poly_architecture(self, output="", scale=10, deformation=[], colors=[]):
        '''
        dump PDB file and tcl script loadable in VMD showing the loaded polyhedral scaffold.

        pseudoatoms are placed in vertices, cylinders connect them. Cylinder color code matches connection type.

        :param scale: vertices scaling factor (i.e. how much you want to blow up your architecture)
        :param colors: list of colors to be used when coloring the cylinders in VMD session. By default, the following 25 VMD colors are available (in this order): blue, red, gray, orange, yellow, tan ,silver, green, white, pink, cyan, purple, lime, mauve, ochre, iceblue, black, yellow2, green2, cyan2, blue2, violet, magenta, red2, orange2.
        :param deformation: if provided, deformations will be applied as described in deformation database
        :param output: name of output files (without extension. .pdb and .tcl will be automatically added). By default, the name will be the polyhedron name.
        '''

        if output == "":
            output = self.polyname

        if len(colors) == 0:
            colors = ['blue', 'red', 'gray', 'orange', 'yellow', 'tan', 'silver', 'green', 'white', 'pink', 'cyan', 'purple', 'lime',
                      'mauve', 'ochre', 'iceblue', 'black', 'yellow2', 'green2', 'cyan2', 'blue2', 'violet', 'magenta', 'red2', 'orange2']

        pos = self.v * scale

        if len(deformation) > 0:
            if len(self.deform) == 0:
                raise Exception("ERROR: %s deformation coefficients provided, but no deformation axis found!" % len(deformation))

            elif len(deformation) == len(self.deform[:, 1]):
                for d in self.deform:
                    pos[int(d[0])] += deformation[int(d[1])] * d[2:5]

            else:
                raise Exception("ERROR: %s deformation coefficients expected, but %s found" % (len(deformation), np.unique(self.deform[:, 1])))

        # output vertices coordinates
        S = Structure(pos)
        S.write_pdb("%s.pdb" % output)

        # output VMD script loading the structure and drawing colored cylinders
        # between them if a connection exists
        fout = open("%s.tcl" % output, "w")
        fout.write("mol new %s.pdb\n" % output)
        fout.write("mol modstyle 0 top VDW 1.000000 12.000000\n")

        for k in range(0, len(self.conn), 1):

            i = pos[self.conn[k][0]]
            j = pos[self.conn[k][1]]

            clr = colors[self.conn_type[k % len(colors)]]
            fout.write("draw color %s\n" % clr)
            fout.write("draw cylinder {%s %s %s} {%s %s %s} radius 0.5\n" %(i[0], i[1], i[2], j[0], j[1], j[2]))

        fout.close()

    def get_neighbors(self, return_chain_names=False):
        '''
        create dictionary defining neighborhood relationships, i.e. edges having a vertex in common (key=edge chain name, value: numpy array of neighboring chains)

        :param return_chain_names: if False, neighbors indices are returned, if True indices will be converted in chain names (as assigned when a pdb is written)
        :returns: dictionary containing neighborhood information
        '''

        # create dictionary with neighbors (key: chain name, values: numpy
        # array of neighboring chains)
        neigh_dict = {}
        for i in range(0, len(self.conn), 1):
            neighs = []
            for j in range(0, len(self.conn), 1):
                if i != j:
                    if self.conn[i, 0] in self.conn[j] or self.conn[i, 1] in self.conn[j]:
                        if return_chain_names:
                            neighs.append(self.chain_names[j])
                        else:
                            neighs.append(j)

            if return_chain_names:
                neigh_dict[self.chain_names[i]] = np.array(neighs)
            else:
                neigh_dict[i] = np.array(neighs)

            return neigh_dict

    def _search_database(self, polyname, dbfilename="polyhedron_database_complete.dat"):
        '''
        search new style database (with plain formatting and containing connectivity information)
        '''

        # check polyhedra database existence
        if os.path.exists(dbfilename) == 0:
            raise Exception('ERROR: file %s not found!' % dbfilename)

        # look for desired polyhedron in database
        fin = open(dbfilename, 'r')
        v = []
        conn = []
        conn_type = []
        for line in fin.readlines():

            linetosave = line.split()
            if linetosave[0] == polyname:
                vert = int(linetosave[1])
                edges = int(linetosave[2])

                # prepare vertices list
                for i in range(vert):
                    v.append((float(linetosave[3 + i * 3]), float(linetosave[3 + i * 3 + 1]), float(linetosave[3 + i * 3 + 2])))

                # prepare connectivity list (could add feature that
                # connectivity is computed, if does not exist)
                for i in range(edges):
                    conn.append((int(linetosave[3 + 3 * vert + i * 2]), int(linetosave[3 + 3 * vert + i * 2 + 1])))

                # prepare connection type information
                if len(linetosave) == 3 + 3 * vert + 2 * edges:
                    print("WARNING: database contains no connection type for %s. Supposing all edges have same connection type." % polyname)
                    conn_type = np.zeros(edges)

                elif len(linetosave) == 3 + 3 * vert + 3 * edges:
                    for i in range(edges):
                        conn_type.append(int(linetosave[3 + 3 * vert + 2 * edges + i]))

                else:
                    raise Exception("ERROR: database inconsistency for connectivity type information in polyhedron %s" % polyname)

                break

        fin.close()

        if len(v) == 0:
            raise Exception("ERROR: polyhedron %s not found in database %s!" %(polyname, dbfilename))

        return edges, np.array(v), np.array(conn), np.array(conn_type)

    def get_polyhedron_properties(self, S):
        '''
        retrieve polyhedron properties
        '''

        x = self.edges * 1.0  # x=number of edges
        y = 2 * self.edges / len(self.v) * 1.0  # y=average connectvity
        psi = 2 * np.pi * (1 / y - 1 / x)
        phi = 2 * np.pi / y  # average facial angle

        top = 1 - np.cos(psi)  # numerator
        bot = 1 - np.cos(phi)  # denominator
        nu = np.arccos(np.sqrt(top / bot))  # tangent curvature angle

        circumradius = S / (2 * np.sin(nu))  # circumradius
        midradius = circumradius * np.cos(nu)  # mid radius

        return psi, phi, nu, circumradius, midradius

    def _polycalc_core(self, W, L, H, S, nu, phi, alpha, beta, gamma, deformation=[], get_box_edges=False):
        '''
        launch polyhedron assembly
        '''

        kay = (2 * H * np.tan(nu) + W / (np.tan(phi / 2) * np.cos(nu))) / L + 1  # scaling factor for truncation
        scale = kay * S  # factor to adjust polyhedron
        vscale = self.v * scale  # rescale polyhedron

        # if needed, apply edges deformation
        if len(deformation) > 0:
            for d in self.deform:
                vscale[int(d[0])] += deformation[int(d[1])] * d[2:5]

        # build bounding box vertices
        vrecty = self._rectanglify(vscale, L, W, H, kay)

        # iterate over all edges
        v_cubody_data = []

        if get_box_edges:
            boxes_edges = []

        for i in range(len(self.conn)):

            # cube located where the protein will have to be displaced
            vcuby = []

            # dig out all 8 vertices for each new rectanglified edge (4 on each
            # end). Distinguish the 'outer' 4 vertices from the 'inner' 4
            # vertices
            for j in range(len(vrecty)):
                if(self.conn[i, 0] == vrecty[j][0] and self.conn[i, 1] == vrecty[j][1]):
                    for k in range(4):
                        vcuby.append((vrecty[j][2 + k][0],
                                      vrecty[j][2 + k][1],
                                      vrecty[j][2 + k][2]))
                elif(self.conn[i, 1] == vrecty[j][0] and self.conn[i, 0] == vrecty[j][1]):
                    for k in range(4):
                        vcuby.append((vrecty[j][2 + k][0],
                                      vrecty[j][2 + k][1],
                                      vrecty[j][2 + k][2]))

            if get_box_edges:
                boxes_edges.extend(vcuby)

            # find the angles required to rotate the central pdb file, to align
            # it with the box (previously was also returning vtbox)
            if 'ndarray' not in str(type(alpha)):
                vtdata = self._cuboid_adjust(np.array(vcuby), self.unit[i].get_xyz(), alpha, beta, gamma, (i, 0))
            else:
                conntype = self.conn_type[i]
                vtdata = self._cuboid_adjust(np.array(vcuby), self.unit[i].get_xyz(), alpha[conntype], beta[conntype], gamma[conntype], (i, 0))

            # append vertex array with pdb file
            v_cubody_data.append(vtdata)

        # protein enclosing boxes
        if get_box_edges:
            Str = Structure(np.array(boxes_edges))
            Str.write_pdb("edges.pdb")

        # return atoms coordinates
        return np.array(v_cubody_data).astype(float)

    def _rectanglify(self, vscale, L, W, H, kay):
        '''
        take a vertex list and return a rectanglified vertex list
        '''

        vrecty = []
        for i in range(len(vscale)):  # for each individual vertex position
            v1 = vscale[i]

            # scroll through the other vertices...
            for j in range(len(vscale)):

                #tt = self._conn_find(i,j)
                tst = False
                for k in range(len(self.conn)):
                    if(self.conn[k, 0] == i and self.conn[k, 1] == j) or (self.conn[k, 1] == i and self.conn[k, 0] == j):
                        tst = True
                        break

                if i != j and tst:  # if the two are connected then get to work!

                    v2 = vscale[j]  # get the vector to the second vertex
                    # get normalised edge vector
                    v2n = (v2 - v1) / np.linalg.norm(v2 - v1)
                    v3c = np.cross(v1, v2)  # get perpendicular to edge
                    v3 = v3c / np.linalg.norm(v3c)

                    # walk short distance from corner to rectangle face
                    i2 = v2n * (kay - 1) * L / 2.0
                    i3 = v3 * W / 2.0  # half width step perpendicular to edge

                    temp1 = v1 + i2  # vector to face of rectangle from corner down edge
                    temp2 = temp1 + i3  # vector to side of rectangle face
                    temp3 = temp1 - i3  # vector to other side of rectangle face

                    # vector perpendicular to edge and width
                    v4c = np.cross(v2n, v3)
                    v4 = v4c / np.linalg.norm(v4c)
                    temp4 = temp2 - v4 * H  # step off outside edge by the height #1
                    temp5 = temp3 - v4 * H  # step off outside edge by heigh #2

                    # take the rectangle face, indexed by vertex into recty array
                    vrecty.append([i, j, temp2, temp3, temp4, temp5])

        return vrecty

    def _cuboid_adjust(self, vcuby, v_data, alpha, beta, gamma, ii):
        '''
        take a cuboid and a source cuboid and data file, and output vertex array with data and box transposed into target position
        ii parameter: if second element not 0, indicates the number of elements per ring (second half rotated by 180 degrees)
        '''

        lim = 1E-7

        # move vertices to centre of mass
        com_cuby = np.mean(vcuby, axis=0)
        vt = vcuby - com_cuby

        if ii[1] != 0:
            # for the double rings, work out what the beta=0 position is based
            # on the centre of face vector
            e = np.mean(vt[[2, 3, 6, 7]], axis=0)

            # work out what centre of face vector makes with Z axis
            betaZero = np.arccos(np.dot(e, np.array([0.0, 0.0, 1.0])) / np.linalg.norm(e))
            # correct to make sure beta is offset by this much
            beta = (np.pi / 2.0 - betaZero) + beta

        tusk = 5.0  # These random rotations make sure that we don't get stuck in an initially bad position
        # perform a random rotation 1
        vt = self._poly_rotate(vt, tusk, 0.0, 0.0)
        # perform a random rotation 2
        vt = self._poly_rotate(vt, 0.0, tusk, 0.0)
        # perform a random rotation 3
        vt = self._poly_rotate(vt, 0.0, 0.0, tusk)

        # find angle of (x,y) component of vector from origin to centre of face
        # (0,1,2,3) makes with positive x axis
        e = np.mean(vt[0:4], axis=0)

        if np.abs(e[0]) != 0 and np.abs(
                e[1]) != 0:  # exception in the case where the 'face' vector is NOT along the +/- z axis
            # theta1=vec_ang((e[0],e[1],0.0),(1.0,0.0,0.0))
            e[2] = 0.0
            theta1 = np.arccos(np.dot(e, np.array([1.0, 0.0, 0.0])) / np.linalg.norm(e))

            if e[1] < 0:
                theta1 = theta1 * -1.0
        else:  # in the case where the face vector IS along the +/- z axis
            theta1 = 0.0

        # perform the z rotation
        vt2 = self._poly_rotate(vt, 0.0, 0.0, theta1)

        # find angle of (z, x) component of vector from origin to centre of
        # face (0, 1, 2, 3) makes with positive x axis
        e = np.mean(vt2[0:4], axis=0)

        if e[2] != 0 and e[
                0] != 0:  # exception for when the face vector is along the +/- y axis
            #theta2=vec_ang((e[0], 0.0, e[2]), (1.0, 0.0, 0.0))
            e[1] = 0.0
            theta2 = np.arccos(np.dot(e, np.array([1.0, 0.0, 0.0])) / np.linalg.norm(e))

            if e[2] > 0:
                theta2 = theta2 * -1.0
        # do this if the vector is along the +/- y axis (TETRAHEDRON WANTS
        # ZERO. SO DOES OCTAHEDRON)
        else:
            theta2 = 0.0

        # perform the 'alpha' lateral rotation
        vt2 = self._poly_rotate(vt2, 0.0, theta2, 0.0)

        # find angle of (y, z) compoent of vector from origin to centre of face
        # (0, 1, 2, 3) makes with positive x axis
        e = vt2[0] - vt2[1]

        # perform the z rotation
        #theta3 = vec_ang((0.0, e[1], e[2]), (0.0, 1.0, 0.0))
        e[0] = 0.0
        theta3 = np.arccos(np.dot(e, np.array([0.0, 1.0, 0.0])) / np.linalg.norm(e))

        if e[2] < 0:
            theta3 = theta3 * -1.0
        vt2 = self._poly_rotate(vt2, theta3, 0.0, 0.0)  # perform an x rotation

        # test for convergence problems
        e = vt2[0] - vt2[1]
        if np.abs(e[0]) > lim:
            return -1

        # PERFORM THE REVERSE ROTATIONS AND TRANSLATION
        vrd = v_data
        if ii[0] >= ii[1] and ii[1] != 0:  # DO THIS FOR DOUBLE RINGS
            vrd = self._poly_rotate(vrd, 0.0, np.pi, 0.0)  # 180 about y
            beta = beta * -1.0

        # perform the 'beta' edge rotation
        vrd = self._poly_rotate(vrd, beta, 0.0, 0.0)
        # perform the 'alpha' lateral rotation
        vrd = self._poly_rotate(vrd, 0.0, 0.0, alpha)
        # perform the 'gamma' lateral rotation
        vrd = self._poly_rotate(vrd, 0.0, gamma, 0.0)
        vrd = self._poly_rotate(vrd, -theta3, 0.0, 0.0)  # x rotation by theta3
        vrd = self._poly_rotate(vrd, 0.0, -theta2, 0.0)  # y rotation by theta2
        vrd = self._poly_rotate(vrd, 0.0, 0.0, -theta1)  # z rotation by theta1
        # perform the reverse random rotation 1
        vrd = self._poly_rotate(vrd, 0.0, 0.0, -tusk)
        # perform the reverse random rotation 2
        vrd = self._poly_rotate(vrd, 0.0, -tusk, 0.0)
        # perform the reverse random rotation 3
        vrd = self._poly_rotate(vrd, -tusk, 0.0, 0.0)

        # perform the reverse of the translation on the structure return
        # coordinates
        return vrd + com_cuby

    def _poly_rotate(self, v_in, a, b, c):
        '''
        function of rotating a point about three euler angles
        requires three rotation matricies (for x y and z axis)
        will return vector from com to new position
        1 0 0         #cosb 0 -sinb  #cosc sinc 0
        0 cosa sina   #0 1 0         #-sinc cosc 0
        0 -sina cosa  #sinb 0 cosb   #0 0 1
        convention is X,Y,Z and right handed
        expects angles to be provided in in radians
        '''

        v = np.array(v_in)

        if a != 0:
            m1 = np.array([[1.0, 0.0, 0.0],
                           [0, np.cos(a), np.sin(a)],
                           [0, -np.sin(a), np.cos(a)]])
            v = np.dot(v, m1.T)

        if b != 0:
            m2 = np.array([[np.cos(b), 0, -np.sin(b)],
                           [0.0, 1.0, 0.0],
                           [np.sin(b), 0, np.cos(b)]])
            v = np.dot(v, m2.T)

        if c != 0:
            m3 = np.array([[np.cos(c), np.sin(c), 0],
                           [-np.sin(c), np.cos(c), 0],
                           [0.0, 0.0, 1.0]])
            return np.dot(v, m3.T)
        else:
            return v



# Copyright (c) 2014-2022 Matteo Degiacomi
#
# BiobOx is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# BiobOx is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with BiobOx ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
#
# Author : Matteo Degiacomi, matteo.degiacomi@gmail.com

'''
Functions to measure characteristics of any Biobox object
'''

import subprocess
import os
import sys
import random
import string

import numpy as np
from ctypes import cdll, c_int, c_float, byref

import biobox.lib.fastmath as FM  # cython routines


def sasa_c(M, targets=[], probe=1.4, n_sphere_point=960, threshold=0.05):
    '''
    compute the accessible surface area using the Shrake-Rupley algorithm ("rolling ball method")

    :param M: any biobox object
    :param targets: indices to be used for surface estimation. By default, all indices are kept into account.
    :param probe: radius of the "rolling ball"
    :param n_sphere_point: number of mesh points per atom
    :param threshold: fraction of points in sphere, above which structure points are considered as exposed
    :returns: accessible surface area in A^2
    :returns: mesh numpy array containing the found points forming the accessible surface mesh
    :returns: IDs of surface points
    '''

    #make sure that everything is collected as a Structure object, and radii are available
    this_inst = type(M).__name__
    if this_inst == "Multimer":
        M = M.make_molecule()

    elif this_inst in ["Assembly", "Polyhedra"]:
        M = M.make_structure()

    # getting radii associated to every atom
    radii = M.data['radius'].values

    if threshold < 0.0 or threshold > 1.0:
        raise Exception("ERROR: threshold should be a floating point between 0 and 1!")

    if len(targets) == 0:
        return FM.c_get_surface(M.points, radii, probe, n_sphere_point, threshold)
    else:
        return FM.c_get_surface(M.points[targets], radii, probe, n_sphere_point, threshold)

def sasa(M, targets=[], probe=1.4, n_sphere_point=960, threshold=0.05):
    '''
    compute the accessible surface area using the Shrake-Rupley algorithm ("rolling ball method")

    :param M: any biobox object
    :param targets: indices to be used for surface estimation. By default, all indices are kept into account.
    :param probe: radius of the "rolling ball"
    :param n_sphere_point: number of mesh points per atom
    :param threshold: fraction of points in sphere, above which structure points are considered as exposed
    :returns: accessible surface area in A^2
    :returns: mesh numpy array containing the found points forming the accessible surface mesh
    :returns: IDs of surface points
    '''

    import biobox.measures.interaction as I

    #make sure that everything is collected as a Structure object, and radii are available
    this_inst = type(M).__name__
    if this_inst == "Multimer":
        M = M.make_molecule()

    elif this_inst in ["Assembly", "Polyhedra"]:
        M = M.make_structure()

    if len(targets) == 0:
        targets = range(0, len(M.points), 1)

    # getting radii associated to every atom
    radii = M.data['radius'].values

    if threshold < 0.0 or threshold > 1.0:
        raise Exception("ERROR: threshold should be a floating point between 0 and 1!")

    # create unit sphere points cloud (using golden spiral)
    pts = []
    inc = np.pi * (3 - np.sqrt(5))
    offset = 2 / float(n_sphere_point)
    for k in range(int(n_sphere_point)):
        y = k * offset - 1 + (offset / 2)
        r = np.sqrt(1 - y * y)
        phi = k * inc
        pts.append([np.cos(phi) * r, y, np.sin(phi) * r])

    sphere_points = np.array(pts)
    const = 4.0 * np.pi / len(sphere_points)

    contact_map = I.distance_matrix(M.points, M.points)

    asa = 0.0
    surface_atoms = []
    mesh_pts = []
    # compute accessible surface for every atom
    for i in targets:

        # place mesh points around atom of choice
        mesh = sphere_points * (radii[i] + probe) + M.points[i]

        # compute distance matrix between mesh points and neighboring atoms
        test = np.where(contact_map[i, :] < radii.max() + probe * 2)[0]
        neigh = M.points[test]
        dist = I.distance_matrix(neigh, mesh) - radii[test][:, np.newaxis]

        # lines=atoms, columns=mesh points. Count columns containing values greater than probe*2
        # i.e. allowing sufficient space for a probe to fit completely
        cnt = 0
        for m in range(dist.shape[1]):
            if not np.any(dist[:, m] < probe):
                cnt += 1
                mesh_pts.append(mesh[m])

        # calculate asa for current atom, if a sufficient amount of mesh
        # points is exposed (NOTE: to verify)
        if cnt > n_sphere_point * threshold:
            surface_atoms.append(i)
            asa += const * cnt * (radii[i] + probe)**2

    return asa, np.array(mesh_pts), np.array(surface_atoms)

def rgyr(M):
    '''
    compute radius of gyration.

    :param M: any biobox object
    :returns: radius of gyration
    '''

    #make sure that everything is collected as a Structure object, and radii are available
    this_inst = type(M).__name__
    if this_inst == "Multimer":
        M = M.make_molecule()

    elif this_inst in ["Assembly", "Polyhedra"]:
        M = M.make_structure()

    d_square = np.sum((M.points - M.get_center())**2, axis=1)
    return np.sqrt(np.sum(d_square) / d_square.shape[0])


def saxs(M, crysol_path='', crysol_options="-lm 20 -ns 500", pdbname=""):
    '''
    compute SAXS curve using crysol (from ATSAS suite)

    :param M: any biobox object
    :param crysol_path: path to crysol executable. If not provided, the environment variable ATSASPATH is sought instead. This allows redirecting to a specific ATSAS bin folder.
    :param crysol_options: flags to be passes to impact executable
    :param pdbname: if a file has been already written, crysol can be asked to analyze it
    :returns: SAXS curve (nx2 numpy array)
    '''

    if crysol_path == '':
        try:
            crysol_path = os.environ['ATSASPATH']
        except KeyError:
            raise Exception("ATSASPATH environment variable undefined")

    if pdbname == "":
        # write temporary pdb file of current structure on which to launch
        # SAXS calculation
        pdbname = "%s.pdb" % random_string(32)
        while os.path.exists(pdbname):
            pdbname = "%s.pdb" % random_string(32)

        M.write_pdb(pdbname, [M.current])

    else:
        # if file was already provided, verify its existence first!
        if os.path.isfile(pdbname) != 1:
            raise Exception("ERROR: %s not found!" % pdbname)

    # get basename for output
    outfile = os.path.basename(pdbname).split('.')[0]

    call_line = os.path.join(crysol_path, "crysol")
    #try:
    subprocess.check_call('%s %s %s > /dev/null' %(call_line, crysol_options, pdbname), shell=True)
    #except Exception as e:
    #    raise Exception("ERROR: crysol calculation failed!")

    data = np.loadtxt("%s00.int" % outfile, skiprows=1)
    try:
        os.remove("%s00.alm" % outfile)
        os.remove("%s00.int" % outfile)
        os.remove("%s00.log" % outfile)
        os.remove("%s.pdb" % outfile)
    except Exception:
        pass

    return data[:, 0:2]


def ccs(M, use_lib=True, impact_path='', impact_options="-Octree -nRuns 32 -cMode sem -convergence 0.01", pdbname="", tjm_scale=False, proberad=1.0):
    '''
    compute CCS calling either impact.

    :param M: any biobox object
    :param use_lib: if true, impact library will be used, if false a system call to impact executable will be performed instead
    :param impact_path: by default, the environment variable IMPACTPATH is sought. This allows redirecting to a specific impact root folder.
    :param impact_options: flags to be passes to impact executable
    :param pdbname: if a file has been already written, impact can be asked to analyze it
    :param tjm_scale: if True, CCS value calculated with PA method is scaled to better match trajectory method.
    :param proberad: radius of probe. Do find out if your impact library already adds this value by default or not (old ones do)!
    :returns: CCS value in A^2. Error return: -1 = input filename not found, -2 = unknown code for CCS calculation\n
              -3 CCS calculator failed, -4 = parsing of CCS calculation results failed
    '''

    #make sure that everything is collected as a Structure object, and radii are available
    this_inst = type(M).__name__
    if this_inst == "Multimer":
        M = M.make_molecule()
        M.assign_atomtype()
        M.get_atoms_ccs()

    elif this_inst in ["Assembly", "Polyhedra"]:
        M = M.make_structure()

    elif this_inst == "Molecule" and "atom_ccs" not in M.data.columns:
        M.assign_atomtype()
        M.get_atoms_ccs()

    if use_lib and pdbname == "":

        #if True:
        from biobox.measures.calculators import CCS
        try:
            if impact_path == '':

                try:
                    impact_path = os.path.join(os.environ['IMPACTPATH'], "lib")
                except KeyError:
                    raise Exception("IMPACTPATH environment variable undefined")

            if "win" in sys.platform:
                libfile = os.path.join(impact_path, "libimpact.dll")
            else:
                libfile = os.path.join(impact_path, "libimpact.so")

            C = CCS(libfile=libfile)

        except Exception as e:
            raise Exception(str(e))

        if "atom_ccs" in M.data.columns:
            radii = M.data['atom_ccs'].values + proberad
        else:
            radii = M.data['radius'].values + proberad

        if tjm_scale:
            return C.get_ccs(M.points, radii)[0]
        else:
            return C.get_ccs(M.points, radii, a=1.0, b=1.0)[0]

    # generate random file name to capture CCS software terminal output
    tmp_outfile = random_string(32)
    while os.path.exists(tmp_outfile):
        tmp_outfile = "%s.pdb" % random_string(32)

    if pdbname == "":
        # write temporary pdb file of current structure on which to launch
        # CCS calculation
        filename = "%s.pdb" % random_string(32)
        while os.path.exists(filename):
            filename = "%s.pdb" % random_string(32)

        M.write_pdb(filename, [M.current])

    else:
        filename = pdbname
        # if file was already provided, verify its existence first!
        if os.path.isfile(pdbname) != 1:
            raise Exception("ERROR: %s not found!" % pdbname)

    try:

        if impact_path == '':
                try:
                    impact_path = os.path.join(os.environ['IMPACTPATH'], "bin")
                except KeyError:
                    raise Exception("IMPACTPATH environment variable undefined")

        # if using impact, create parameterization file containing a
        # description for Z atoms (pseudoatom name used in this code)
        f = open('params', 'w')

        f.write('[ defaults ]\n H 2.2\n C 2.91\n N 2.91\n O 2.91\n P 2.91\n S 2.91\n')
        #@fix this for the general case of multiple atoms with different radius
        f.write(' Z %s' % (np.unique(M.data['radius'])[0] + proberad))
        impact_options += " -param params"

        f.close()

        if "win" in sys.platform:
            impact_name = os.path.join(impact_path, "impact.exe")
        else:
            impact_name = os.path.join(impact_path, "impact")

        subprocess.check_call('%s  %s -rProbe 0 %s > %s' % (impact_name, impact_options, filename, tmp_outfile), shell=True)

    except Exception as e:
        raise Exception(str(e))

    #parse output generated by IMPACT and written into a file
    try:
        f = open(tmp_outfile, 'r')
        for line in f:
            w = line.split()
            if len(w) > 0 and w[0] == "CCS":

                if tjm_scale:
                    v = float(w[-2])
                else:
                    v = float(w[3])

                break

        f.close()

        # clean temp files if needed
        #(if a filename is provided, don't delete it!)
        os.remove(tmp_outfile)
        if pdbname == "":
            os.remove(filename)

        return v

    except:
        # clean temp files
        os.remove(tmp_outfile)
        if pdbname == "":
            os.remove(filename)

        return -4



class CCS(object):
    '''
    CCS calculator (wrapper for C library)
    '''

    def __init__(self, libfile):
        '''
        initialize by loading IMPACT library

        :param libfile: library path
        '''

        try:
            self.libs = cdll.LoadLibrary(libfile)
            self.libs.pa2tjm.restype = c_float

        except:
            raise Exception("loading library %s failed!" % libfile)

        # declare output variables
        self.ccs = c_float()
        self.sem = c_float()
        self.niter = c_int()

    def get_ccs(self, points, radii, a=0.842611, b=1.051280):
        '''
        compute CCS using the PA method as implemented in IMPACT library.

        :param points: xyz coordinates of atoms, Angstrom (nx3 numpy array)
        :param radii: van der Waals radii associated to every point (numpy array with n elements)
        :param a: power-law factor for calibration with TJM
        :param b: power-law exponent for calibration with TJM
        :returns: TJM CCS
        :returns: standard error
        :returns: number of iterations
        '''

        # create ctypes for intput data
        unravel = np.ravel(points)
        cpoints = (c_float * len(unravel))(*unravel)
        cradii = (c_float * len(radii))(*radii)
        natoms = (c_int)(len(radii))

        # call library, and rescale obtained value using exponential law
        self.libs.ccs_from_atoms_defaults(natoms, byref(cpoints), byref(cradii), byref(self.ccs), byref(self.sem), byref(self.niter))
        ccs_tjm = self.libs.pa2tjm(c_float(a), c_float(b), self.ccs)

        return ccs_tjm, self.sem.value, self.niter.value


def random_string(length=32):
    '''
    generate a random string of arbitrary characters. Useful to generate temporary file names.

    :param length: length of random string
    '''
    return ''.join([random.choice(string.ascii_letters)
                    for n in range(length)])



# Copyright (c) 2014-2022 Matteo Degiacomi
#
# BiobOx is free software ;
# you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation ;
# either version 2 of the License, or (at your option) any later version.
# BiobOx is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY ;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with BiobOx ;
# if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
#
# Author : Matteo Degiacomi, matteo.degiacomi@gmail.com

from copy import deepcopy
import numpy as np
import scipy.signal
import pandas as pd

class Structure(object):
    '''
    A Structure consists of an ensemble of points in 3D space, and metadata associated to each of them.
    '''

    def __init__(self, p=np.array([[], []]), r=1.0):
        '''
        Point coordinates and properties data structures are first initialized.
        properties is a dictionary initially containing an entry for 'center' (center of geometry) and 'radius' (average radius of points).

        :param p: coordinates data structure as a mxnx3 numpy array (alternative conformation x atom x 3D coordinate). nx3 numpy array can be supplied, in case a single conformation is present.
        :param r: average radius of every point in dataset (float), or radius of every point (numpy array)
        '''
        if p.ndim == 3:
            self.coordinates = p
            '''numpy array containing an ensemble of alternative coordinates in 3D space'''

        elif p.ndim == 2:
            self.coordinates = np.array([p])
        else:
            raise Exception("ERROR: expected numpy array with 2 or three dimensions, but %s dimensions were found" %p.ndim)

        self.current = 0
        '''index of currently selected conformation'''

        #self.points = self.coordinates[self.current]
        self.points = self.coordinates.view()[self.current]
        '''pointer to currently selected conformation'''

        self.properties = {}
        '''collection of properties. By default, 'center' (geometric center of the Structure) is defined'''

        self.properties['center'] = self.get_center()

        idx = np.arange(len(self.points))
        if isinstance(r, list) or type(r).__module__ == 'numpy':
            if len(r) > 0:
                self.data = pd.DataFrame(r, index=idx, columns=["radius"])
        else:
                rad = r*np.ones(len(self.points))
                self.data = pd.DataFrame(rad, index=idx, columns=["radius"])
                ''' metadata about each atom (pandas Dataframe)'''

    def __len__(self, dim="atoms"):
        if dim == "atoms":
            return len(self.points)

    def __getitem__(self, key):
        return self.coordinates[key]

    def set_current(self, pos):
        '''
        select current frame (place frame pointer at desired position)

        :param pos: number of alternative conformation (starting from 0)
        '''
        if pos < self.coordinates.shape[0]:
            self.current = pos
            #self.points = self.coordinates[self.current]
            self.points = self.coordinates.view()[self.current]
            self.properties['center'] = self.get_center()
        else:
            raise Exception("ERROR: position %s requested, but only %s conformations available" %(pos, self.coordinates.shape[0]))

    def get_xyz(self, indices=[]):
        '''
        get points coordinates.

        :param indices: indices of points to select. If none is provided, all points coordinates are returned.
        :returns: coordinates of all points indexed by the provided indices list, or all of them if no list is provided.
        '''
        if indices == []:
            return self.points
        else:
            return self.points[indices]

    def set_xyz(self, coords):
        '''
        set point coordinates.

        :param coords: array of 3D points
        '''
        self.coordinates[self.current] = deepcopy(coords)
        #self.points = self.coordinates[self.current]
        self.points = self.coordinates.view()[self.current]

    def add_xyz(self, coords):
        '''
        add a new alternative conformation to the database

        :param coords: array of 3D points, or array of arrays of 3D points (in case multiple alternative coordinates must be added at the same time)
        '''
        # self.coordinates numpy array containing an ensemble of alternative
        # coordinates in 3D space

        if self.coordinates.size == 0 and coords.ndim == 3:
            self.coordinates = deepcopy(coords)
            self.set_current(0)

        elif self.coordinates.size == 0 and coords.ndim == 2:
            self.coordinates = deepcopy(np.array([coords]))
            self.set_current(0)

        elif self.coordinates.size > 0 and coords.ndim == 3:
            self.coordinates = np.concatenate((self.coordinates, coords))
            # set new frame to the first of the newly inserted ones
            self.set_current(self.current + 1)

        elif self.coordinates.size > 0 and coords.ndim == 2:
            self.coordinates = np.concatenate((self.coordinates, np.array([coords])))
            # set new frame to the first of the newly inserted ones
            self.set_current(self.current + 1)

        else:
            raise Exception("ERROR: expected numpy array with 2 or three dimensions, but %s dimensions were found" %np.ndim)


    def delete_xyz(self, index):
        '''
        remove one conformation from the conformations database.

        the new current conformation will be the previous one.

        :param index: alternative coordinates set to remove
        '''
        self.coordinates = np.delete(self.coordinates, index, axis=0)
        if index > 0:
            self.set_current(index - 1)
        else:
            self.set_current(0)

    def clear(self):
        '''
        remove all the coordinates and empty metadata
        '''
        self.coordinates = np.array([[[], []], [[], []]])
        #self.points = self.coordinates[0]
        self.points = self.coordinates.view()[0]
        self.data = pd.DataFrame(index=[], columns=[])

    def translate(self, x, y, z):
        '''
        translate the whole structure by a given amount.

        :param x: translation around x axis
        :param y: translation around y axis
        :param z: translation around z axis
        '''

        # if center has not been defined yet (may happen when using
        # subclasses), compute it
        if 'center' not in self.properties:
            self.get_center()

        # translate all points
        self.properties['center'][0] += x
        self.properties['center'][1] += y
        self.properties['center'][2] += z

        # move every frame with first :
        self.coordinates[:, :, 0] += x
        self.coordinates[:, :, 1] += y
        self.coordinates[:, :, 2] += z

    def rotate(self, x, y, z):
        '''
        rotate the structure provided angles of rotation around x, y and z axes (in degrees).

        This is a rotation with respect of the origin.
        Make sure that the center of your structure is at the origin, if you don't want to get a translation as well!
        rotating an object being not centered requires to first translate the ellipsoid at the origin, rotate it, and bringing it back.

        :param x: rotation around x axis
        :param y: rotation around y axis
        :param z: rotation around z axis
        '''
        alpha = np.radians(x)
        beta = np.radians(y)
        gamma = np.radians(z)
        Rx = np.array([[1, 0, 0],
                       [0, np.cos(alpha), - np.sin(alpha)],
                       [0, np.sin(alpha), np.cos(alpha)]])
        Ry = np.array([[np.cos(beta), 0, np.sin(beta)],
                       [0, 1, 0],
                       [-np.sin(beta), 0, np.cos(beta)]])
        Rz = np.array([[np.cos(gamma), -np.sin(gamma), 0],
                       [np.sin(gamma), np.cos(gamma), 0],
                       [0, 0, 1]])
        rotation = np.dot(Rx, np.dot(Ry, Rz))
        # multiply rotation matrix with each point of the ellipsoid
        self.apply_transformation(rotation.T)
        self.get_center()

    def apply_transformation(self, M):
        '''
        apply a 3x3 transformation matrix

        :param M: 3x3 transformation matrix (2D numpy array)
        '''
        self.coordinates[self.current, :, :] = np.dot(self.points, M)
        # new memory allocated? Pointer needs to be moved
        #self.points = self.coordinates[self.current]
        self.points = self.coordinates.view()[self.current]

    def get_center(self):
        '''
        compute protein center of geometry (also assigns it to self.properties["center"] key).
        '''
        if len(self.points) > 0:
            self.properties['center'] = np.mean(self.points, axis=0)
        else:
            self.properties['center'] = np.array([0.0, 0.0, 0.0])

        return self.properties['center']

    def center_to_origin(self):
        '''
        move the structure so that its center of geometry is at the origin.
        '''
        c = self.get_center()
        self.translate(-c[0], -c[1], -c[2])

    def get_size(self):
        '''
        compute the dimensions of the object along x, y and z.

        .. note: points radii are not kept into account.
        '''
        x = np.max(self.points[:, 0]) - np.min(self.points[:, 0])
        # +self.properties['radius']*2
        y = np.max(self.points[:, 1]) - np.min(self.points[:, 1])
        # +self.properties['radius']*2
        z = np.max(self.points[:, 2]) - np.min(self.points[:, 2])
        return np.array([x, y, z])

    def rotation_matrix(self, axis, theta):
        '''
        compute matrix needed to rotate the system around an arbitrary axis (using Euler-Rodrigues formula).

        :param axis: 3d vector (numpy array), representing the axis around which to rotate
        :param theta: desired rotation angle
        :returns: 3x3 rotation matrix
        '''

        # if rotation angle is equal to zero, no rotation is needed
        if theta == 0:
            return np.identity(3)

        # method taken from
        # http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector
        axis = axis / np.sqrt(np.dot(axis, axis))
        a = np.cos(theta / 2)
        b, c, d = -axis * np.sin(theta / 2)
        return np.array([[a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)],
                         [2 * (b * c + a * d), a * a + c * c - b * b - d * d, 2 * (c * d - a * b)],
                         [2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c]])

    def get_principal_axes(self):
        '''
        compute Structure's principal axes.

        :returns: 3x3 numpy array, containing the 3 principal axes ranked from smallest to biggest.
        '''
        # method taken from chempy source code, geometry.py, method
        # getMomentOfInertiaTensor()

        # compute moment of inertia tensor
        I0 = np.zeros((3, 3), np.float64)
        for i in range(0, len(self.points), 1):
            mass = 1  # self.mass[atom] / constants.Na
            I0[0, 0] += mass * (self.points[i, 1] * self.points[i, 1] + self.points[i, 2] * self.points[i, 2])
            I0[1, 1] += mass * (self.points[i, 0] * self.points[i, 0] + self.points[i, 2] * self.points[i, 2])
            I0[2, 2] += mass * (self.points[i, 0] * self.points[i, 0] + self.points[i, 1] * self.points[i, 1])
            I0[0, 1] -= mass * self.points[i, 0] * self.points[i, 1]
            I0[0, 2] -= mass * self.points[i, 0] * self.points[i, 2]
            I0[1, 2] -= mass * self.points[i, 1] * self.points[i, 2]

        I0[1, 0] = I0[0, 1]
        I0[2, 0] = I0[0, 2]
        I0[2, 1] = I0[1, 2]

        # Calculate and return the principal moments of inertia and corresponding
        # principal axes for the current geometry.
        e_values, e_vectors = np.linalg.eig(I0)

        indices = np.argsort(e_values)
        e_values = e_values[indices]
        e_vectors = e_vectors.T[indices]

        return e_vectors

    def align_axes(self):
        '''
        Align structure on its principal axes.

        First principal axis aligned along x, second along y and third along z.
        '''

        # this method is inspired from the procedure followed in in VMD's orient package:
        # set I [draw principalaxes $sel]           <--- show/calc the principal axes
        # set A [orient $sel [lindex $I 2] {0 0 1}] <--- rotate axis 2 to match Z
        # $sel move $A
        # set I [draw principalaxes $sel]           <--- recalc principal axes to check
        # set A [orient $sel [lindex $I 1] {0 1 0}] <--- rotate axis 1 to match Y
        # $sel move $A
        # set I [draw principalaxes $sel]           <--- recalc principal axes
        # to check

        # this align axes has been modified to allow us to backmap a structure after alignment
        c = self.get_center()

        # center the Structure
        self.center_to_origin()

        # get principal axes (ranked from smallest to biggest)
        axes = self.get_principal_axes()

        # align smallest principal axis against z axis
        rotvec = np.cross(axes[0], np.array([1, 0, 0]))  # rotation axis
        sine = np.linalg.norm(rotvec)
        cosine = np.dot(axes[0], np.array([1, 0, 0]))
        angle = np.arctan2(sine, cosine)  # angle to rotate around axis

        rotmatrix0 = self.rotation_matrix(rotvec, angle)
        self.apply_transformation(rotmatrix0)

        # compute new principal axes (after previous rotation)
        axes = self.get_principal_axes()

        # align second principal axis against y axis
        rotvec = np.cross(axes[1], np.array([0, 1, 0]))  # rotation axis
        sine = np.linalg.norm(rotvec)
        cosine = np.dot(axes[1], np.array([0, 1, 0]))
        angle = np.arctan2(sine, cosine)  # angle to rotate around axis

        rotmatrix1 = self.rotation_matrix(rotvec, angle)
        self.apply_transformation(rotmatrix1)

        # return the center and matrix for backmapping
        # do the opposite of these transformations
        return c, rotmatrix0, rotmatrix1

    def write_pdb(self, filename, index=[]):
        '''
        write a multi PDB file where every point is a sphere. VdW radius is written into beta factor.

        :param filename: name of file to output
        :param index: list of frame indices to write to file. By default, a multipdb with all frames will be produced.
        '''

        # if a subset of all available frames is requested to be written,
        # select them first
        if len(index) == 0:
            frames = range(0, len(self.coordinates), 1)
        else:
            if np.max(index) < len(self.coordinates):
                frames = index
            else:
                raise Exception("ERROR: requested coordinate index %s, but only %s are available" %(np.max(index), len(self.coordinates)))

        fout = open(filename, "w")

        for f in frames:

            # Build our hexidecimal array if num. of atoms > 99999
            idx_val = np.arange(1, self.coordinates.shape[1] + 1, 1)

            if len(idx_val) > 99999:
                vhex = np.vectorize(hex)
                idx_val = vhex(idx_val)   # convert index values to hexidecimal
                idx_val = [num[2:] for num in idx_val]  # remove 0x at start of hexidecimal number

            for i in range(0, len(self.coordinates[0]), 1):
                #if i > 99999:
                #    nb = hex(i).split('x')[1]
                #else:
                #    nb = str(i)

                l = (idx_val[i], "SPH", "SPH", "A", np.mod(i, 9999),
                     self.coordinates[f, i, 0],
                     self.coordinates[f, i, 1],
                     self.coordinates[f, i, 2],
                     self.data['radius'].values[i],
                     1.0, "Z")
                L = 'ATOM  %5s  %-4s%-4s%1s%4i    %8.3f%8.3f%8.3f%6.2f%6.2f          %2s\n' % l
                fout.write(L)

            fout.write("END\n")

        fout.close()

    def convex_hull(self):
        '''
        compute Structure's convex Hull using QuickHull algorithm.

        .. note:: Qhull available only on scipy >=0.12

        :returns: :func:`Structure <structure.Structure>` object, containing the coordinates of vertices composing the convex hull
        '''
        try:
            from scipy.spatial import ConvexHull
            verts = ConvexHull(self.points)
            return Structure(verts)

        except Exception as e:
            raise Exception("Quick Hull algorithm available in scipy >=0.12!")

    def get_density(self, step=1.0, sigma=1.0, kernel_half_width=5, buff=3):
        '''
        generate density map from points

        :param step: size of cubic voxels, in Angstrom
        :param sigma: gaussian kernel sigma
        :param kernel_half_width: kernel half width, in voxels
        :param buff: padding to add at points cloud boundaries
        :returns: :func:`Density <density.Density>` object, containing a simulated density map
        '''
        pts = self.points

        # rectangular box boundaries
        bnds = np.array([[np.min(pts[:, 0]) - buff, np.max(pts[:, 0]) + buff],
                         [np.min(pts[:, 1]) - buff, np.max(pts[:, 1]) + buff],
                         [np.min(pts[:, 2]) - buff, np.max(pts[:, 2]) + buff]])

        xax = np.arange(bnds[0, 0], bnds[0, 1] + step, step)
        yax = np.arange(bnds[1, 0], bnds[1, 1] + step, step)
        zax = np.arange(bnds[2, 0], bnds[2, 1] + step, step)

        # create empty box
        d = np.zeros((len(xax), len(yax), len(zax)))

        # place Kronecker deltas in mesh grid
        for p in pts:
            xpos = np.argmin(np.abs(xax - p[0]))
            ypos = np.argmin(np.abs(yax - p[1]))
            zpos = np.argmin(np.abs(zax - p[2]))
            d[xpos, ypos, zpos] = 1

        # create 3d gaussian kernel
        window = kernel_half_width * 2 + 1
        shape = (window, window, window)

        m, n, k = [(ss - 1.) / 2. for ss in shape]

        x_ = np.arange(-m, m + 1, 1).astype(int)
        y_ = np.arange(-n, n + 1, 1).astype(int)
        z_ = np.arange(-k, k + 1, 1).astype(int)
        x, y, z = np.meshgrid(x_, y_, z_)

        h = np.exp(-(x * x + y * y + z * z) / (2. * sigma * sigma))
        h[h < np.finfo(h.dtype).eps * h.max()] = 0
        sumh = h.sum()
        if sumh != 0:
            h /= sumh

        # convolve point mesh with 3d gaussian kernel
        b = scipy.signal.fftconvolve(d, h, mode='same')
        b /= np.max(b)

        # prepare density data structure
        from biobox.classes.density import Density
        D = Density()
        D.properties['density'] = b
        D.properties['size'] = np.array(b.shape)
        D.properties['origin'] = np.min(self.points, axis=0) - kernel_half_width / 2.0 + step #np.mean(self.points, axis=0) - step * np.array(b.shape) / 2.0
        D.properties['delta'] = np.identity(3) * step
        D.properties['format'] = 'dx'
        D.properties['filename'] = ''
        D.properties["sigma"] = np.std(b)

        return D

    def rmsf(self, indices=-1, step=1):
        '''
        compute Root Mean Square Fluctuation (RMSF) of selected atoms.

        :param indices: indices of points for which RMSF will be calculated. If no indices list is provided, RMSF of all points will be calculated.
        :param step: timestep between two conformations (useful when using conformations extracted from molecular dynamics)
        :returns: numpy aray with RMSF of all provided indices, in the same order
        '''

        if self.coordinates.shape[0] < 2:
            raise Exception("ERROR: to compute RMSF several conformations must be available!")

        # if no index is provided, compute RMSF of all points
        if indices == -1:
            indices = np.linspace(0, len(self.coordinates[0, :, 0]) - 1, len(self.coordinates[0, :, 0])).astype(int)

        means = np.mean(self.coordinates[:, indices], axis=0)

        # cumulate all squared distances with respect of mean
        d = []
        for i in range(0, self.coordinates.shape[0], 1):
            d.append(np.sum((self.coordinates[i, indices] - means)**2, axis=1))

        # compute square root of sum of mean squared distances
        dist = np.array(d)
        return np.sqrt(np.sum(dist, axis=0) / (float(self.coordinates.shape[0]) * step))

    def pca(self, components, indices=-1):
        '''
        compute Principal Components Analysis (PCA) on specific points within all the alternative coordinates.

        :param components: eigenspace dimensions
        :param indices: points indices to be considered for PCA
        :returns: numpy array of projection of each conformation into the n-dimensional eigenspace
        :returns: sklearn PCA object
        '''

        from sklearn.decomposition import PCA

        # define conformational space (flatten coordinates of desired atoms
        if indices != -1:
            X = self.coordinates[:, indices].reshape(
                     (len(self.coordinates), len(indices) * 3))
        else:
            X = self.coordinates.reshape(
                     (self.coordinates.shape[0], self.coordinates.shape[1]*3))

        # calculate system PCA and project conformations into the eigenspace
        pca = PCA(n_components=components)
        pca.fit(X)
        Xproj = pca.transform(X)

        return Xproj, pca

    def rmsd_one_vs_all(self, ref_index, points_index=[], align=False):
        '''
        Calculate the RMSD between all structures with respect of a reference structure.
        uses Kabsch alignement algorithm.

        :param ref_index: index of reference structure in conformations database
        :param points_index: if set, only specific points will be considered for comparison
        :param align: if set to true, all conformations will be aligned to reference (note: cannot be undone!)
        :returns: RMSD of all structures with respect of reference structure (in a numpy array)
        '''

        # see: http://www.pymolwiki.org/index.php/Kabsch#The_Code

        bkpcurrent = self.current

        if ref_index >= len(self.coordinates):
            raise Exception("ERROR: index %s requested, but only %s exist in database" %(len(self.coordinates)))

        # define reference frame, and center it
        if len(points_index) == 0:
            m1 = deepcopy(self.coordinates[ref_index])
        elif isinstance(points_index, list) or type(points_index).__module__ == 'numpy':
            m1 = deepcopy(self.coordinates[ref_index, points_index])
        else:
            raise Exception("ERROR: please, provide me with a list of indices to compute RMSD (or no index at all)")

        L = len(m1)
        COM1 = np.sum(m1, axis=0) / float(L)
        m1 -= COM1
        m1sum = np.sum(np.sum(m1 * m1, axis=0), axis=0)

        RMSD = []
        for i in range(0, len(self.coordinates), 1):

            if i == ref_index:
                RMSD.append(0.0)
            else:

                # define current frame, and center it
                if len(points_index) == 0:
                    m2 = deepcopy(self.coordinates[i])
                elif isinstance(points_index, list) or type(points_index).__module__ == 'numpy':
                    m2 = deepcopy(self.coordinates[i, points_index])

                COM2 = np.sum(m2, axis=0) / float(L)
                m2 -= COM2

                E0 = m1sum + np.sum(np.sum(m2 * m2, axis=0), axis=0)

                # This beautiful step provides the answer. V and Wt are the orthonormal
                # bases that when multiplied by each other give us the rotation matrix, U.
                # S, (Sigma, from SVD) provides us with the error!  Isn't SVD
                # great!
                V, S, Wt = np.linalg.svd(np.dot(np.transpose(m2), m1))

                # if alignement is required, move pointer to current frame, and
                # apply rotation matrix
                if align:
                    self.set_current(i)
                    rotation = np.dot(V, Wt)
                    self.apply_transformation(rotation)

                reflect = float(
                    str(float(np.linalg.det(V) * np.linalg.det(Wt))))

                if reflect == -1.0:
                    S[-1] = -S[-1]
                    V[:, -1] = -V[:, -1]

                rmsdval = E0 - (2.0 * sum(S))
                rmsdval = np.sqrt(abs(rmsdval / L))

                RMSD.append(rmsdval)

        self.set_current(bkpcurrent)
        return np.array(RMSD)

    def rmsd(self, i, j, points_index=[], full=False):
        '''
        Calculate the RMSD between two structures in alternative coordinates ensemble.
        uses Kabsch alignement algorithm.

        :param i: index of the first structure
        :param j: index of the second structure
        :param points_index: if set, only specific points will be considered for comparison
        :param full: if True, RMSD an rotation matrx are returned, RMSD only otherwise
        :returns: RMSD of the two structures. If full is True, the rotation matrix is also returned
        '''

        # see: http://www.pymolwiki.org/index.php/Kabsch#The_Code

        if i >= len(self.coordinates):
            raise Exception("ERROR: index %s requested, but only %s exist in database" %(i, len(self.coordinates)))

        if j >= len(self.coordinates):
            raise Exception("ERROR: index %s requested, but only %s exist in database" %(j, len(self.coordinates)))

        # get first structure and center it
        if len(points_index) == 0:
            m1 = deepcopy(self.coordinates[i])
        elif isinstance(points_index, list) or type(points_index).__module__ == 'numpy':
            m1 = deepcopy(self.coordinates[i, points_index])
        else:
            raise Exception("ERROR: give me a list of indices to compute RMSD, or nothing at all, please!")

        # get second structure
        if len(points_index) == 0:
            m2 = deepcopy(self.coordinates[j])
        elif isinstance(points_index, list) or type(points_index).__module__ == 'numpy':
            m2 = deepcopy(self.coordinates[j, points_index])
        else:
            raise Exception("ERROR: give me a list of indices to compute RMSD, or nothing at all, please!")

        L = len(m1)
        COM1 = np.sum(m1, axis=0) / float(L)
        m1 -= COM1
        m1sum = np.sum(np.sum(m1 * m1, axis=0), axis=0)

        COM2 = np.sum(m2, axis=0) / float(L)
        m2 -= COM2

        E0 = m1sum + np.sum(np.sum(m2 * m2, axis=0), axis=0)

        # This beautiful step provides the answer. V and Wt are the orthonormal
        # bases that when multiplied by each other give us the rotation matrix, U.
        # S, (Sigma, from SVD) provides us with the error!  Isn't SVD great!
        V, S, Wt = np.linalg.svd(np.dot(np.transpose(m2), m1))

        reflect = float(str(float(np.linalg.det(V) * np.linalg.det(Wt))))

        if reflect == -1.0:
            S[-1] = -S[-1]
            V[:, -1] = -V[:, -1]

        rmsdval = E0 - (2.0 * sum(S))
        if full:
            return np.sqrt(abs(rmsdval / L)), np.matmul(V, Wt)
        else:
            return np.sqrt(abs(rmsdval / L))

    def rmsd_distance_matrix(self, points_index=[], flat=False):
        '''
        compute distance matrix between structures (using RMSD as metric).

        :param points_index: if set, only specific points will be considered for comparison
        :param flat: if True, returns flattened distance matrix
        :returns: RMSD distance matrix
        '''

        if flat:
            rmsd = []
        else:
            rmsd = np.zeros((len(self.coordinates), len(self.coordinates)))

        for i in range(0, len(self.coordinates) - 1, 1):
            for j in range(i + 1, len(self.coordinates), 1):
                r = self.rmsd(i, j, points_index)

                if flat:
                    rmsd.append(r)
                else:
                    rmsd[i, j] = r
                    rmsd[j, i] = r

        if flat:
            return np.array(rmsd)
        else:
            return rmsd



import unittest
import sys, os
if 'CONDA_BUILD_STATE' in os.environ and os.environ['CONDA_BUILD_STATE']=='TEST':
    pass
else:
    sys.path.append(os.sep.join(os.getcwd().split(os.sep)[:-1])+os.sep+'src')
import biobox as bb

class test_density(unittest.TestCase):

    def setUp(self):
        self.D = bb.Density()
        self.D.import_map("EMD-1080.mrc", "mrc")

    def test_density_points(self):

        print("\n> density: placing points")

        try:
            self.D.place_points(5)
        except Exception:
            assert False

    def test_density_CCS(self):

        print("\n> density: CCS calculation")

        try:
            self.D.threshold_vol_ccs(sampling_points=1, append=False, noise_filter=0)
        except Exception:
            assert False


class test_structures(unittest.TestCase):

    def setUp(self):
        self.M = bb.Molecule()
        self.M.import_pdb("HSP.pdb")

    def test_len(self):
        print("\n> testing magic method:")
        print(">> atom count of monomer: %s"%len(self.M))

        M2 = self.M + self.M
        print(">> atom count of dimer: %s"%len(M2))
        #print M2[:, 20:22]

    def test_xlink(self):

        print("\n> testing shortest path")

        try:
            #extract indices of atoms to connect
            idx = self.M.atomselect("*", "LYS", "NZ", use_resname=True, get_index=True)[1]

            #prepare xlink measurer
            XL = bb.Xlink(self.M)
            XL.set_clashing_atoms(atoms=["CA", "C", "N", "O", "CB"], densify=True, atoms_vdw=False)

            XL.setup_global_search(maxdist=14, use_hull=False)
            #XL.setup_local_search(maxdist=24)

            distance2, paths = XL.distance_matrix(idx, method="theta", get_path=True, smooth=True, verbose=False, flexible_sidechain=True, test_los=True)

        except Exception:
            assert False


    def test_SASA(self):

        print("\n> testing molecule's SASA")
        try:
            [sasa, mesh, surf_idx] = bb.sasa(self.M, n_sphere_point=400)
        except Exception:
            assert False


    def test_monomer_CCS(self):
        if 'IMPACTPATH' in os.environ:
            print("\n> testing CCS")
        else:
            print("\n\n> IMPACTPATH not set therefore can't test CCS. \n WARNING: If you want CCS calculations you need to set IMPACTPATH")
            return

        try:
            ccs1 = bb.ccs(self.M)
        except Exception:
            assert False

        try:
            ccs2 = bb.ccs(self.M, use_lib=False)
        except Exception:
            assert False

        self.assertAlmostEqual(ccs1, ccs2, delta=ccs2/10.0) #max 10% difference


    def test_multimer_CCS(self):
        if 'IMPACTPATH' in os.environ:
            print("\n> testing multimer CCS")
        else:
            print("\n\n> IMPACTPATH not set therefore can't test CCS. \n WARNING: If you want CCS calculations you need to set IMPACTPATH")
            return


        try:
            A = bb.Multimer()
            A.load(self.M, 3)
            A.make_circular_symmetry(30)
            bb.ccs(A)
        except Exception:
            assert False

    def test_multimer_selections(self):

        print("\n> testing multimer atomselect and query")
        try:
            A = bb.Multimer()
            A.load_list([self.M, self.M], ["1", "2"])
        except Exception:
            assert False

        pts_test = self.M.atomselect("*", "LYS", "CA", use_resname = True)

        try:
            pts = A.query('unit == "1" and resname == "LYS" and name == "CA"')
        except Exception:
            assert False

        self.assertEqual(len(pts), len(pts_test))

        try:
            pts = A.atomselect("1", "*", "LYS", "CA", use_resname = True)
        except Exception:
            assert False

        self.assertEqual(len(pts), len(pts_test))


    def test_multimer_rototranslations(self):

        print("\n> testing multimer rototranslations")
        try:
            P = bb.Multimer()
            P.load(self.M, 6)
            P.rotate(0, 0, 90)
            P.make_prism(25, 15, 180, 45, 90)
            P.rotate(10, 10, 10, [1, 2])
        except Exception:
            assert False


    #test rototranslations on double disks (prism method)
    def test_monomers_rototranslations(self):

        print("\n> testing monomer rototranslations")
        try:
            self.M.align_axes()
            self.M.rotate(10, 10, 10)
            self.M.translate(20, 20, 20)

        except Exception:
            assert False


    #test assembly of multiple polyhedral architectures, and RMSD evaluation
    def test_polyRMSD(self):

        print("\n> assemblying Polyhedra")
        try:
            #setup desired polyhedron
            P = bb.Multimer()
            P.setup_polyhedron("Octahedron", self.M)

            #try creation and and deletion of some polyhedra
            P.generate_polyhedron(40, 180, 0, 0)
            P.generate_polyhedron(42, 180, 5, 0, add_conformation=True)
            P.generate_polyhedron(40, 200, 5, 0, add_conformation=True)
            P.generate_polyhedron(40, 180, 10, 10, add_conformation=True)
            P.generate_polyhedron(40, 180, 5, 5, add_conformation=True)
            P.delete_xyz(2)

            #test atomselects on alternate conformations
            P.set_current(0)
            a1 = P.atomselect(["1", "2"], "*", 90, "CA")
            P.set_current(1)
            a2 = P.atomselect(["1", "2"], "*", 90, "CA")

            #different conformations must be selected
            self.assertNotAlmostEqual(a1[0, 0], a2[0, 0])

        except Exception:
            assert False


    #create all convex shapes
    def test_shapes(self):

        print("\n> testing convex shapes")
        try:
            C1 = bb.Prism(10, 20, 5)
            C1.get_surface()
            C1.get_volume()

            C2 = bb.Cylinder(10, 50)
            C2.get_surface()
            C2.get_volume()

            C3 = bb.Cone(10, 30)
            C3.get_surface()
            C3.get_volume()

            C4 = bb.Ellipsoid(10, 20, 30)
            C4.get_surface()
            C4.get_volume()

            C5 = bb.Sphere(10)
            C5.get_surface()
            C5.get_volume()

        except Exception:
            assert False


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import numpy as np
from distutils.core import setup
from distutils.command.build_ext import build_ext
from Cython.Build import cythonize

class InstallCommand(build_ext):

    def run(self):

        build_ext.run(self)

        try:
            for root, dirnames, filenames in os.walk("build"):
                for filename in filenames:
                    extension = filename.split(".")[-1]
                    if extension in ["pyd", "so"]:
                        os.rename(os.path.join(root, filename), filename)

        except Exception as ex:
            print("files already exist, skipping...")

        shutil.rmtree("build")

# small hack to get around a problem in older cython versions, i.e.
# an infinite dependencies loop when __init__.py file is in the same folder as pyx
if os.path.exists("__init__.py"):
    os.rename("__init__.py", "tmp")


setup(
    include_dirs=[np.get_include()],
    ext_modules=cythonize(
        "e_density.pyx",
        include_path=[np.get_include()],
            compiler_directives={'boundscheck': False, 'wraparound': False}),
    cmdclass={'install': InstallCommand}
)

# continuation of the small hack
os.rename("tmp", "__init__.py")

